import configparser
from copy import deepcopy


class Config(object):
    """
        This class stores and manipulates a configuration file.

        The initialization of the class will execute the load method.

        ...

        Parameters
        ----------
        fullpath : str
            path to the file.
        validators : dict
            dict where the keys are the name of the sections and the values
            are another dict.
            In the latter dict, the keys are the name of the options and the values
            are methods to convert the raw information (e.g: int)

        Attributes
        ----------

        fullpath : str
            path to the file
        config : dict
            validated values of each option of each section
        configparser : configparser.ConfigParser object
            ConfigParser object with raw values (str format) of each option of each section
        validators : dict
            dict of validator methods for each section and option (default is {})
        sections : list
            list of the section names

        Methods
        -------
        preview()
            print the text of the file.

        load()
            load the file located at .fullpath

        save(fullpath=None)
            save the current configuration into a file. (default is None)
            If it is None, it will use the current .fullpath

        get_options(section)
            get the list of option names of a given section

        config_to_configparser()
            copy the information in .config to .configparser

        set(section, option, value)
            change the value of a given option of a section in .config
            it applies the validator if there is any

        get(section, option)
            return the value of a given option of a section in .config

        add_section(section)
            create a new section in the configuration

    """
    def __init__(self, fullpath, validators=None):
        if validators is None:
            validators = {}
        self.configparser = configparser.ConfigParser()
        self.validators = validators
        if fullpath:
            self.fullpath = fullpath
            self.load()

    def load(self):
        self.read_file()
        self.configparser_to_config()
        self.validate_config()

    def read_file(self):
        self.configparser.read(self.fullpath)

    def configparser_to_config(self):
        self.config = deepcopy(self.configparser._sections)

    def validate_config(self):
        for section in self.sections:
            if section not in self.validators.keys():
                continue
            for option, value in self.config[section].items():
                if option not in self.validators[section].keys():
                    continue
                validated_value = self.validators[section][option](value)
                self.config[section][option] = validated_value

    @property
    def sections(self):
        return self.config.keys()

    def preview(self):
        for sec in self.sections:
            print('[%s]' % (sec))
            for key, value in self.config[sec].items():
                print('{} = {}'.format(key, value))
            print('')

    def get_options(self, section):
        options = self.config[section]
        return options

    def save(self, fullpath=None):
        self.config_to_configparser()
        if fullpath is None:
            fullpath = self.fullpath
        else:
            self.fullpath = fullpath
        with open(self.fullpath, 'w') as file:
            self.configparser.write(file)

    def config_to_configparser(self):
        for sec in self.sections:
            for key, value in self.config[sec].items():
                self.configparser[sec][key] = str(value)

    def set(self, section, option, value):
        if section in self.validators.keys():
            if option in self.validators[section].keys():
                value = self.validators[section][option](value)
        self.config[section][option] = value

    def get(self, section, option):
        return self.config[section][option]

    def add_section(self, section):
        if section not in self.sections:
            self.config[section] = {}
            self.configparser.add_section(section)

    def __repr__(self):
        lines = []
        lines.append(f'{self.__class__.__name__}({self.fullpath})')
        for sec in self.sections:
            lines.append(f'[{sec}]')
            for key, value in self.config[sec].items():
                lines.append(f'{key} = {value}')
        out = '\n'.join(lines)
        return out

    def __str__(self):
        return self.__repr__()



import shapely
import shapely.geometry
import shapely.ops
import matplotlib.pyplot as plt
import numpy as np


class SlicedPolygon(shapely.geometry.Polygon):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._slices = []
        self.slice()

    @property
    def bounds_x(self):
        return np.array([self.bounds[0], self.bounds[2]])

    @property
    def bounds_y(self):
        return np.array([self.bounds[1], self.bounds[3]])

    @property
    def x(self):
        return np.array(self.boundary.xy[0])

    @property
    def y(self):
        return np.array(self.boundary.xy[1])

    @property
    def x_unique(self):
        return np.unique(self.x)

    @property
    def y_unique(self):
        return np.unique(self.y)

    @property
    def slice_axis(self):
        if len(self.x_unique) <= len(self.y_unique):
            axis = 'x'
        else:
            axis = 'y'
        return axis

    @property
    def slice_points(self):
        if self.slice_axis == 'x':
            arr = self.x_unique
        elif self.slice_axis == 'y':
            arr = self.y_unique
        points = np.copy(arr[1:-1])
        return points

    @property
    def sliceable(self):
        boolean = True
        if not list(self.slice_points):
            boolean = False
        return boolean

    @property
    def slice_line(self):
        if self.slice_axis == 'x':
            index = 0
            bounds = self.bounds_y
        else:
            index = 1
            bounds = self.bounds_x
        points = []
        for i, si in enumerate(self.slice_points):
            # -- For each slice point, we need to create 2 points
            point = [0, 0]
            point[index] = si
            point[(index + 1) % 2] = bounds[i % 2]
            points.append(list(point))
            point[index] = si
            point[(index + 1) % 2] = bounds[(i + 1) % 2]
            points.append(list(point))
        return shapely.geometry.LineString(points)

    @property
    def slices(self):
        return self._slices

    def slice(self):
        if not self.sliceable:
            slices = [self]
        else:
            slices = shapely.ops.split(self, self.slice_line)
        self._slices = slices
        return slices

    def correct_xy(self):
        xs, ys = [], []
        for polygon in self.slices:
            x, y = np.array(polygon.boundary.xy)
            x, y = self._remove_redundant_vertexes(x, y)
            x, y = self._correct_xy_for_triangles(x, y)
            x, y = self._order_xy_for_nn(x, y)
            xs.append(x)
            ys.append(y)
        return xs, ys

    def _remove_redundant_vertexes(self, x, y, remove_last_point=True):
        xy = np.array([x, y])
        xy = np.transpose(xy)
        if remove_last_point and (xy[0] == xy[-1]).all():  # remove the last point (repeated)
            xy = np.delete(xy, -1, axis=0)
        idx_to_delete = []
        nb_points = xy.shape[0]
        for i in range(nb_points):
            i_l = i - 1
            i_m = i
            i_r = int((i + 1) % nb_points)
            xy_l = xy[i_l]
            xy_m = xy[i_m]
            xy_r = xy[i_r]
            v1 = xy_m - xy_l
            v2 = xy_r - xy_m
            if np.cross(v1, v2) == 0:
                idx_to_delete.append(i)
        # -- Delete indexes
        # -- Note that you need to delete them in reverse order so that you don't throw off the subsequent indexes.
        for ind in sorted(idx_to_delete, reverse=True):
            xy = np.delete(xy, ind, axis=0)
        xy = np.transpose(xy)
        x, y = xy
        return x, y

    def _correct_xy_for_triangles(self, x, y):
        nb_vertexes = len(x)
        if nb_vertexes == 3:  # it is a triangle
            if self.slice_axis == 'x':
                p_check = x
            elif self.slice_axis == 'y':
                p_check = y

            if p_check[0] == p_check[1]:
                duplicate_index = 2
            elif p_check[0] == p_check[2]:
                duplicate_index = 1
            else:
                duplicate_index = 0

            x = np.append(x, x[duplicate_index])
            y = np.append(y, y[duplicate_index])
        return x, y

    def _order_xy_for_nn(self, x, y):
        if self.slice_axis == 'x':
            p_sort = x
            p_follow = y
        elif self.slice_axis == 'y':
            p_sort = y
            p_follow = x

        p_sorted = np.sort(p_sort)
        arg_sorted = np.argsort(p_sort)
        p_follow = p_follow[arg_sorted]

        if self.slice_axis == 'x':
            x = p_sorted
            y = p_follow
        elif self.slice_axis == 'y':
            y = p_sorted
            x = p_follow
        return x, y

    def show_slices(self, ax=None, fill_kw={}):
        if not ax:
            fig, ax = plt.subplots(1)
        ax.set_xlabel(f'x')
        ax.set_ylabel(f'y')
        for pol in self.slices:
            x, y = pol.boundary.xy
            ax.fill(x, y, **fill_kw)
        return ax


import unittest
import warnings

import nextnanopy.outputs as outputs
from nextnanopy.utils.datasets import default_unit
from pathlib import Path

folder_nnp = Path("tests") / "datafiles" / "nextnano++"
folder_nn3 = Path("tests") / "datafiles" / "nextnano3"
folder_negf = Path("tests") / "datafiles" / "nextnano.NEGF"
folder_msb = Path("tests") / "datafiles" / "nextnano.MSB"


class TestOutputs_nnp(unittest.TestCase):

    def test_dat(self):
        df = outputs.DataFile(
            folder_nnp / "bandedges_1d.dat", product="nextnano++"
        )
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords["x"].name, "x")
        self.assertEqual(df.coords["x"].unit, "nm")
        self.assertEqual(df.coords["x"].value.size, 548)
        self.assertEqual(len(df.variables.keys()), 4)
        self.assertEqual(df.variables["Gamma"].name, "Gamma")
        self.assertEqual(df.variables["Gamma"].unit, "eV")
        self.assertEqual(df.variables["Gamma"].value.size, 548)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

        df = outputs.DataFile(
            folder_nnp / "wf_occupation_1d.dat", product="nextnano++"
        )
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.coords["no."].name, "no.")
        self.assertEqual(df.coords["no."].unit, default_unit)
        self.assertEqual(df.coords["no."].value.size, 10)
        self.assertEqual(df.variables["Occupation"].name, "Occupation")
        self.assertEqual(df.variables["Occupation"].unit, "electrons/cm^2")
        self.assertEqual(df.variables["Occupation"].value.size, 10)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

    def test_dat_FirstVarFlag(self):
        df = outputs.DataFile(
            folder_nnp / "bandedges_1d.dat",
            product="nextnano++",
            FirstVarIsCoordFlag=False,
        )
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 5)
        self.assertEqual(df.variables["x"].name, "x")
        self.assertEqual(df.variables["x"].unit, "nm")
        self.assertEqual(df.variables["x"].value.size, 548)
        self.assertEqual(df.variables["Gamma"].name, "Gamma")
        self.assertEqual(df.variables["Gamma"].unit, "eV")
        self.assertEqual(df.variables["Gamma"].value.size, 548)
        self.assertEqual(df.metadata["ndim"], 0)
        self.assertEqual(df.metadata["dkeys"], [])

        df = outputs.DataFile(
            folder_nnp / "wf_occupation_1d.dat",
            product="nextnano++",
            FirstVarIsCoordFlag=True,
        )
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.coords["no."].name, "no.")
        self.assertEqual(df.coords["no."].unit, default_unit)
        self.assertEqual(df.coords["no."].value.size, 10)
        self.assertEqual(df.variables["Occupation"].name, "Occupation")
        self.assertEqual(df.variables["Occupation"].unit, "electrons/cm^2")
        self.assertEqual(df.variables["Occupation"].value.size, 10)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

    # test creating the document
    # self.assertRaises(NotImplementedError, df.export, filename='TESTEXPORT.dat', format='dat')

    def test_avs(self):
        df = outputs.DataFile(
            folder_nnp / "bandedges_2d.fld", product="nextnano++"
        )
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ["x", "y"])
        self.assertEqual(df.coords["x"].unit, "nm")
        self.assertEqual(df.coords["x"].value.size, 164)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, "nm")
        self.assertEqual(df.coords["y"].value.size, 79)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(len(df.variables.keys()), 2)
        self.assertEqual(df.variables["Gamma"].name, "Gamma")
        self.assertEqual(df.variables["Gamma"].unit, "eV")
        self.assertEqual(df.variables["Gamma"].value.shape, (164, 79))
        self.assertEqual(df.metadata["ndim"], 2)
        self.assertEqual(df.metadata["dims"], [164, 79])
        self.assertEqual(df.metadata["field"], "rectilinear")

        # test creating the document
        self.assertRaises(
            NotImplementedError,
            df.export,
            filename="TESTEXPORT2D.fld",
            format="AVSAscii",
        )

    def test_avs3D(self):
        df = outputs.DataFile(folder_nnp / "potential.fld", product="nextnano++")
        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ["x", "y", "z"])
        self.assertEqual(df.coords["x"].unit, "nm")
        self.assertEqual(df.coords["x"].value.size, 11)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, "nm")
        self.assertEqual(df.coords["y"].value.size, 11)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(df.coords["z"].unit, "nm")
        self.assertEqual(df.coords["z"].value.size, 76)
        self.assertEqual(df.coords["z"].dim, 2)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables["potential"].name, "potential")
        self.assertEqual(df.variables["potential"].unit, default_unit)
        self.assertEqual(df.variables["potential"].value.shape, (11, 11, 76))
        self.assertEqual(df.metadata["ndim"], 3)
        self.assertEqual(df.metadata["dims"], [11, 11, 76])
        self.assertEqual(df.metadata["field"], "rectilinear")

        # test creating the document
        self.assertRaises(
            NotImplementedError,
            df.export,
            filename="TESTEXPORT3D.fld",
            format="AVSAscii",
        )

    def test_vtr_potential(self):
        df = outputs.DataFile(folder_nnp / "potential.vtr", product="nextnano++")

        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ["x", "y", "z"])
        self.assertEqual(df.coords["x"].unit, default_unit)
        self.assertEqual(df.coords["x"].value.size, 11)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, default_unit)
        self.assertEqual(df.coords["y"].value.size, 11)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(df.coords["z"].unit, default_unit)
        self.assertEqual(df.coords["z"].value.size, 76)
        self.assertEqual(df.coords["z"].dim, 2)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables["potential"].name, "potential")
        self.assertEqual(df.variables["potential"].unit, default_unit)
        self.assertEqual(df.variables["potential"].value.shape, (11, 11, 76))
        self.assertEqual(df.metadata, {})

    def test_vtr_bandedges(self):
        df = outputs.DataFile(folder_nnp / "bandedges.vtr", product="nextnano++")

        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ["x", "y", "z"])
        self.assertEqual(df.coords["x"].unit, default_unit)
        self.assertEqual(df.coords["x"].value.size, 11)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, default_unit)
        self.assertEqual(df.coords["y"].value.size, 11)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(df.coords["z"].unit, default_unit)
        self.assertEqual(df.coords["z"].value.size, 76)
        self.assertEqual(df.coords["z"].dim, 2)
        self.assertEqual(len(df.variables.keys()), 4)
        self.assertEqual(df.variables["Gamma"].name, "Gamma")
        self.assertEqual(df.variables["Gamma"].unit, "eV")
        self.assertEqual(df.variables["Gamma"].value.shape, (11, 11, 76))
        self.assertEqual(df.metadata, {})

    def test_txt(self):
        df = outputs.DataFile(
            folder_nnp / "variables_input.txt", product="nextnano++"
        )
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 3)
        self.assertEqual(df.variables["var_1"].name, "var_1")
        self.assertEqual(df.variables["var_1"].unit, default_unit)
        self.assertEqual(df.variables["var_1"].value, -47.5)
        self.assertEqual(df.variables["var_2"].value, 75)
        self.assertEqual(df.variables["string"].value, '"hello"')

        df = outputs.DataFile(
            folder_nnp / "variables_database.txt", product="nextnano++"
        )
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 0)

        df = outputs.DataFile(
            folder_nnp / "total_charges.txt", product="nextnano++"
        )
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 6)
        self.assertAlmostEqual(df.variables["fixed"].value, 0.0)
        self.assertEqual(df.variables[0].name, "electron")
        self.assertEqual(df.variables[0].unit, "e/cm^2")

        df = outputs.DataFile(folder_nnp / "overlap_integrals_k00000.txt", product="nextnano++")

        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 5)
        self.assertAlmostEqual(df.variables["|Gamma_i>"].value[0], 1)
        self.assertAlmostEqual(df.variables["<Gamma_i|HH_j>"].value[0], 0.0333932)




    def test_rest(self):
        files = ["example.log", "example.in"]
        for file in files:
            self.assertRaises(
                NotImplementedError,
                outputs.DataFile,
                folder_nnp / file,
                "nextnano++",
            )

    def test_all_2D(self):
        files = [
            Path("AvsAscii") / "bandedges.avs.fld",
            Path("AvsAscii_one_file") / "bandedges.fld",
            Path("AvsBinary") / "bandedges.avs.fld",
            Path("AvsBinary_one_file") / "bandedges.fld",
            Path("VTKAscii") / "bandedges.vtr",
        ]
        for file in files:
            filepath = folder_nnp / file
            datafile = outputs.DataFile(filepath, product="nextnano++")
            self.assertEqual(len(datafile.coords), 2)
            self.assertEqual(len(datafile.variables), 6)

    def test_binary_ascii_mix(self):
        file = Path("AvsBinaryAscii_mix") / "ldos_total_cbr_Gamma.fld"
        filepath = folder_nnp / file
        datafile = outputs.DataFile(filepath, product="nextnano++")
        self.assertEqual(len(datafile.coords), 2)
        self.assertEqual(len(datafile.variables), 1)
    
    # TODO add tests for loading Origin like file
    # TODO add tests for exporting binary data


class TestOutputs_nn3(unittest.TestCase):

    def test_dat(self):
        df = outputs.DataFile(folder_nn3 / "bandedges_1d.dat", product="nextnano3")
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords["position"].name, "position")
        self.assertEqual(df.coords["position"].unit, "nm")
        self.assertEqual(df.coords["position"].value.size, 258)
        self.assertEqual(len(df.variables.keys()), 6)
        self.assertEqual(df.variables["Gamma_bandedge"].name, "Gamma_bandedge")
        self.assertEqual(df.variables["Gamma_bandedge"].unit, "eV")
        self.assertEqual(df.variables["Gamma_bandedge"].value.size, 258)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

        df = outputs.DataFile(folder_nn3 / "wf_shift_1d.dat", product="nextnano3")
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(len(df.variables.keys()), 10)
        self.assertEqual(df.coords["position"].name, "position")
        self.assertEqual(df.coords["position"].unit, "nm")
        self.assertEqual(df.coords["position"].value.size, 121)
        self.assertEqual(df.variables["psi^2_1"].name, "psi^2_1")
        self.assertEqual(df.variables["psi^2_1"].unit, default_unit)
        self.assertEqual(df.variables["psi^2_1"].value.size, 121)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

        df = outputs.DataFile(
            folder_nn3 / "bandedges_2d_cut.dat", product="nextnano3"
        )
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords["position"].name, "position")
        self.assertEqual(df.coords["position"].unit, "nm")
        self.assertEqual(df.coords["position"].value.size, 258)
        self.assertEqual(len(df.variables.keys()), 6)
        self.assertEqual(df.variables["Gamma_bandedge"].name, "Gamma_bandedge")
        self.assertEqual(df.variables["Gamma_bandedge"].unit, "eV")
        self.assertEqual(df.variables["Gamma_bandedge"].value.size, 258)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

    def test_dat_FirstVarIsCoord(self):
        df = outputs.DataFile(
            folder_nn3 / "bandedges_1d.dat",
            product="nextnano3",
            FirstVarIsCoordFlag=False,
        )
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 7)
        self.assertEqual(df.variables["position"].name, "position")
        self.assertEqual(df.variables["position"].unit, "nm")
        self.assertEqual(df.variables["position"].value.size, 258)
        self.assertEqual(df.variables["Gamma_bandedge"].name, "Gamma_bandedge")
        self.assertEqual(df.variables["Gamma_bandedge"].unit, "eV")
        self.assertEqual(df.variables["Gamma_bandedge"].value.size, 258)
        self.assertEqual(df.metadata["ndim"], 0)
        self.assertEqual(df.metadata["dkeys"], [])

        df = outputs.DataFile(
            folder_nn3 / "wf_shift_1d.dat",
            product="nextnano3",
            FirstVarIsCoordFlag=True,
        )
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(len(df.variables.keys()), 10)
        self.assertEqual(df.coords["position"].name, "position")
        self.assertEqual(df.coords["position"].unit, "nm")
        self.assertEqual(df.coords["position"].value.size, 121)
        self.assertEqual(df.variables["psi^2_1"].name, "psi^2_1")
        self.assertEqual(df.variables["psi^2_1"].unit, default_unit)
        self.assertEqual(df.variables["psi^2_1"].value.size, 121)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

    def test_avs1D(self):
        df = outputs.DataFile(folder_nn3 / "cb_Gamma_avs.fld", product="nextnano3")
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(list(df.coords.keys()), ["x"])
        self.assertEqual(df.coords["x"].unit, "nm")
        self.assertEqual(df.coords["x"].value.size, 105)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables["cb_Gamma"].name, "cb_Gamma")
        self.assertEqual(df.variables["cb_Gamma"].unit, "eV")
        self.assertEqual(df.variables["cb_Gamma"].value.shape, (105,))
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dims"], [105])
        self.assertEqual(df.metadata["field"], "rectilinear")

    def test_avs(self):
        df = outputs.DataFile(folder_nn3 / "bandedges_2d.fld", product="nextnano3")
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ["x", "y"])
        self.assertEqual(df.coords["x"].unit, "nm")
        self.assertEqual(df.coords["x"].value.size, 258)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, "nm")
        self.assertEqual(df.coords["y"].value.size, 11)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(len(df.variables.keys()), 6)
        self.assertEqual(df.variables["Gamma_bandedge"].name, "Gamma_bandedge")
        self.assertEqual(df.variables["Gamma_bandedge"].unit, "eV")
        self.assertEqual(df.variables["Gamma_bandedge"].value.shape, (258, 11))
        self.assertEqual(df.metadata["ndim"], 2)
        self.assertEqual(df.metadata["dims"], [258, 11])
        self.assertEqual(df.metadata["field"], "rectilinear")

    def test_avs2D(self):
        df = outputs.DataFile(
            folder_nn3 / "2Dcb1_sg1_deg1_psi_ev001.fld", product="nextnano3"
        )
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ["x", "y"])
        self.assertEqual(df.coords["x"].unit, "nm")
        self.assertEqual(df.coords["x"].value.size, 61)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, "nm")
        self.assertEqual(df.coords["y"].value.size, 61)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(df.variables[0], df.variables["psi_real"])
        self.assertEqual(df.variables[0].name, "psi_real")
        self.assertEqual(df.variables[0].unit, "nm^-1")
        self.assertEqual(df.variables[0].value.shape, (61, 61))
        self.assertEqual(df.variables[1], df.variables["psi_imag"])
        self.assertEqual(df.variables[1].name, "psi_imag")
        self.assertEqual(df.variables[1].unit, "nm^-1")
        self.assertEqual(df.variables[1].value.shape, (61, 61))
        self.assertEqual(df.metadata["ndim"], 2)
        self.assertEqual(df.metadata["dims"], [61, 61])
        self.assertEqual(df.metadata["field"], "rectilinear")

    def test_avs3D(self):
        df = outputs.DataFile(
            folder_nn3 / "3Dcb1_sg1_deg1_psi_squared_ev001.fld",
            product="nextnano3",
        )
        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ["x", "y", "z"])
        self.assertEqual(df.coords["x"].unit, "nm")
        self.assertEqual(df.coords["x"].value.size, 19)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, "nm")
        self.assertEqual(df.coords["y"].value.size, 19)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(df.coords["z"].unit, "nm")
        self.assertEqual(df.coords["z"].value.size, 9)
        self.assertEqual(df.coords["z"].dim, 2)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables["psi_squared"].name, "psi_squared")
        self.assertEqual(df.variables["psi_squared"].unit, "nm^-3")
        self.assertEqual(df.variables[0].value.shape, (19, 19, 9))
        self.assertEqual(df.metadata["ndim"], 3)
        self.assertEqual(df.metadata["dims"], [19, 19, 9])
        self.assertEqual(df.metadata["field"], "rectilinear")

    def test_vtr_LocalDOS(self):
        df = outputs.DataFile(
            folder_nn3 / "LocalDOS_sg1_deg1_Lead1.vtr", product="nextnano3"
        )

        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ["x", "y"])
        self.assertEqual(df.coords["x"].unit, default_unit)
        self.assertEqual(df.coords["x"].value.size, 51)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, default_unit)
        self.assertEqual(df.coords["y"].value.size, 101)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(
            df.variables["LocalDOS_sg1_deg1_Lead1"].name, "LocalDOS_sg1_deg1_Lead1"
        )
        self.assertEqual(df.variables["LocalDOS_sg1_deg1_Lead1"].unit, "nm^-1eV^-1")
        self.assertEqual(df.variables["LocalDOS_sg1_deg1_Lead1"].value.shape, (51, 101))
        self.assertEqual(df.metadata, {})

    def test_vtr_2Dcb1(self):
        df = outputs.DataFile(
            folder_nn3 / "2Dcb1_sg1_deg1_psi_ev001.vtr", product="nextnano3"
        )

        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ["x", "y"])
        self.assertEqual(df.coords["x"].unit, default_unit)
        self.assertEqual(df.coords["x"].value.size, 61)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, default_unit)
        self.assertEqual(df.coords["y"].value.size, 61)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(len(df.variables.keys()), 2)
        self.assertEqual(df.variables[0], df.variables["psi_real"])
        self.assertEqual(df.variables[0].name, "psi_real")
        self.assertEqual(df.variables[0].unit, "nm^-1")
        self.assertEqual(df.variables[0].value.shape, (61, 61))
        self.assertEqual(df.variables[1], df.variables["psi_imag"])
        self.assertEqual(df.variables[1].name, "psi_imag")
        self.assertEqual(df.variables[1].unit, "nm^-1")
        self.assertEqual(df.variables[1].value.shape, (61, 61))
        self.assertEqual(df.metadata, {})

    def test_vtr_3Dcb1(self):
        df = outputs.DataFile(
            folder_nn3 / "3Dcb1_sg1_deg1_psi_squared_ev001.vtr",
            product="nextnano3",
        )

        self.assertEqual(len(df.coords.keys()), 3)
        self.assertEqual(list(df.coords.keys()), ["x", "y", "z"])
        self.assertEqual(df.coords["x"].unit, default_unit)
        self.assertEqual(df.coords["x"].value.size, 19)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, default_unit)
        self.assertEqual(df.coords["y"].value.size, 19)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(df.coords["z"].unit, default_unit)
        self.assertEqual(df.coords["z"].value.size, 9)
        self.assertEqual(df.coords["z"].dim, 2)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables["psi_squared"].name, "psi_squared")
        self.assertEqual(df.variables["psi_squared"].unit, "nm^-3")
        self.assertEqual(df.variables["psi_squared"].value.shape, (19, 19, 9))
        self.assertEqual(df.metadata, {})

    def test_txt(self):
        df = outputs.DataFile(
            folder_nn3 / "variables_input.txt", product="nextnano3"
        )
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 3)
        self.assertEqual(df.variables["var_1"].name, "var_1")
        self.assertEqual(df.variables["var_1"].unit, default_unit)
        self.assertEqual(df.variables["var_1"].value, -47.5)
        self.assertEqual(df.variables["var_2"].value, 75)
        self.assertEqual(df.variables["string"].value, '"hello"')

        df = outputs.DataFile(
            folder_nn3 / "variables_database.txt", product="nextnano3"
        )
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 0)

    def test_rest(self):
        files = ["example.log", "example.in"]
        for file in files:
            self.assertRaises(
                NotImplementedError,
                outputs.DataFile,
                folder_nnp / file,
                "nextnano3",
            )

    # TODO add tests for loading binary (different binary + avs not one file)


class TestOutputs_negf(unittest.TestCase):

    def test_dat(self):
        df = outputs.DataFile(
            folder_negf / "ReducedRealSpaceModes.dat", product="nextnano.NEGF"
        )
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords["Position"], df.coords[0])
        self.assertEqual(df.coords["Position"].name, "Position")
        self.assertEqual(df.coords["Position"].unit, "nm")
        self.assertEqual(df.coords["Position"].value.size, 608)
        self.assertEqual(len(df.variables.keys()), 19)
        self.assertEqual(df.variables["Conduction_BandEdge"], df.variables[0])
        self.assertEqual(
            df.variables["Conduction_BandEdge"].name, "Conduction_BandEdge"
        )
        self.assertEqual(df.variables["Conduction_BandEdge"].unit, "eV")
        self.assertEqual(df.variables["Conduction_BandEdge"].value.size, 608)
        self.assertEqual(df.variables["Psi_1 (lev.1 per.0)"], df.variables[1])
        self.assertEqual(
            df.variables["Psi_1 (lev.1 per.0)"].name, "Psi_1 (lev.1 per.0)"
        )
        self.assertEqual(df.variables["Psi_1 (lev.1 per.0)"].unit, "a.u.")
        self.assertEqual(df.variables["Psi_1 (lev.1 per.0)"].value.size, 608)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

        df = outputs.DataFile(
            folder_negf / "E_p (Kane energy).dat", product="nextnano.NEGF"
        )
        self.assertEqual(df.coords["Position"], df.coords[0])
        self.assertEqual(df.coords["Position"].name, "Position")
        self.assertEqual(df.coords["Position"].unit, "nm")
        self.assertEqual(df.coords["Position"].value.size, 203)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables["E_p"].name, "E_p")
        self.assertEqual(df.variables["E_p"].unit, default_unit)
        self.assertEqual(df.variables["E_p"].value.size, 203)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

    def test_dat_FirstVarIsCoordFlag(self):
        df = outputs.DataFile(
            folder_negf / "ReducedRealSpaceModes.dat",
            product="nextnano.NEGF",
            FirstVarIsCoordFlag=False,
        )
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 20)
        self.assertEqual(df.variables["Position"].name, "Position")
        self.assertEqual(df.variables["Position"].unit, "nm")
        self.assertEqual(df.variables["Position"].value.size, 608)
        self.assertEqual(df.variables["Conduction_BandEdge"], df.variables[1])
        self.assertEqual(
            df.variables["Conduction_BandEdge"].name, "Conduction_BandEdge"
        )
        self.assertEqual(df.variables["Conduction_BandEdge"].unit, "eV")
        self.assertEqual(df.variables["Conduction_BandEdge"].value.size, 608)
        self.assertEqual(df.variables["Psi_1 (lev.1 per.0)"], df.variables[2])
        self.assertEqual(
            df.variables["Psi_1 (lev.1 per.0)"].name, "Psi_1 (lev.1 per.0)"
        )
        self.assertEqual(df.variables["Psi_1 (lev.1 per.0)"].unit, "a.u.")
        self.assertEqual(df.variables["Psi_1 (lev.1 per.0)"].value.size, 608)
        self.assertEqual(df.metadata["ndim"], 0)
        self.assertEqual(df.metadata["dkeys"], [])

        df = outputs.DataFile(
            folder_negf / "E_p (Kane energy).dat",
            product="nextnano.NEGF",
            FirstVarIsCoordFlag=True,
        )
        self.assertEqual(df.coords["Position"], df.coords[0])
        self.assertEqual(df.coords["Position"].name, "Position")
        self.assertEqual(df.coords["Position"].unit, "nm")
        self.assertEqual(df.coords["Position"].value.size, 203)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables["E_p"].name, "E_p")
        self.assertEqual(df.variables["E_p"].unit, default_unit)
        self.assertEqual(df.variables["E_p"].value.size, 203)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

    def test_dat_no_units(self):
        df = outputs.DataFile(
            folder_negf / "datafile_with_no_units.dat", product="nextnano.NEGF"
        )
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords["Photon Energy"], df.coords[0])
        self.assertEqual(df.coords[0].name, "Photon Energy")
        self.assertEqual(df.coords["Photon Energy"].unit, "meV")
        self.assertEqual(df.coords["Photon Energy"].value.size, 3670)
        self.assertEqual(len(df.variables.keys()), 36)
        self.assertEqual(df.variables["Transition 2 => 3"], df.variables[0])
        self.assertEqual(df.variables["Transition 2 => 3"].unit, "")

        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

    def test_vtr(self):
        df = outputs.DataFile(
            folder_negf / "CarrierDensity_energy_resolved.vtr",
            product="nextnano.NEGF",
        )
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ["x", "y"])
        self.assertEqual(df.coords["x"].unit, default_unit)
        self.assertEqual(df.coords["x"].value.size, 176)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, default_unit)
        self.assertEqual(df.coords["y"].value.size, 82)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables["Carrier density"].name, "Carrier density")
        self.assertEqual(df.variables["Carrier density"].unit, "cm^-3 eV^-1")
        self.assertEqual(df.variables["Carrier density"].value.shape, (176, 82))
        self.assertEqual(df.metadata, {})

    def test_rest(self):
        files = [
            "example.log",
            "example.in",
            "Material_Database.xml",
            "Convergence.txt",
        ]
        for file in files:
            self.assertRaises(
                NotImplementedError,
                outputs.DataFile,
                folder_negf / file,
                "nextnano.NEGF",
            )


# NEGF C# and NEGF C++ have identical outputs. Maybe no need to test.


class TestOutputs_msb(unittest.TestCase):

    def test_dat(self):
        df = outputs.DataFile(
            folder_msb / "BandEdge_conduction.dat", product="nextnano.MSB"
        )
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords["Position"].name, "Position")
        self.assertEqual(df.coords["Position"].unit, "nm")
        self.assertEqual(df.coords["Position"].value.size, 100)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(
            df.variables["Conduction Band Edge"].name, "Conduction Band Edge"
        )
        self.assertEqual(df.variables["Conduction Band Edge"].unit, "eV")
        self.assertEqual(df.variables["Conduction Band Edge"].value.size, 100)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

    def test_dat_FirstVarIsCoordFlag(self):
        df = outputs.DataFile(
            folder_msb / "BandEdge_conduction.dat",
            product="nextnano.MSB",
            FirstVarIsCoordFlag=False,
        )
        self.assertEqual(len(df.coords.keys()), 0)
        self.assertEqual(len(df.variables.keys()), 2)
        self.assertEqual(df.variables["Position"].name, "Position")
        self.assertEqual(df.variables["Position"].unit, "nm")
        self.assertEqual(df.variables["Position"].value.size, 100)
        self.assertEqual(
            df.variables["Conduction Band Edge"].name, "Conduction Band Edge"
        )
        self.assertEqual(df.variables["Conduction Band Edge"].unit, "eV")
        self.assertEqual(df.variables["Conduction Band Edge"].value.size, 100)
        self.assertEqual(df.metadata["ndim"], 0)
        self.assertEqual(df.metadata["dkeys"], [])

        df = outputs.DataFile(
            folder_msb / "BandEdge_conduction.dat",
            product="nextnano.MSB",
            FirstVarIsCoordFlag=True,
        )
        self.assertEqual(len(df.coords.keys()), 1)
        self.assertEqual(df.coords["Position"].name, "Position")
        self.assertEqual(df.coords["Position"].unit, "nm")
        self.assertEqual(df.coords["Position"].value.size, 100)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(
            df.variables["Conduction Band Edge"].name, "Conduction Band Edge"
        )
        self.assertEqual(df.variables["Conduction Band Edge"].unit, "eV")
        self.assertEqual(df.variables["Conduction Band Edge"].value.size, 100)
        self.assertEqual(df.metadata["ndim"], 1)
        self.assertEqual(df.metadata["dkeys"], [0])

    def test_vtr(self):
        df = outputs.DataFile(
            folder_msb / "DOS_Lead_Source_position_resolved.vtr",
            product="nextnano.MSB",
        )
        self.assertEqual(len(df.coords.keys()), 2)
        self.assertEqual(list(df.coords.keys()), ["x", "y"])
        self.assertEqual(df.coords["x"].unit, default_unit)
        self.assertEqual(df.coords["x"].value.size, 50)
        self.assertEqual(df.coords["x"].dim, 0)
        self.assertEqual(df.coords["y"].unit, default_unit)
        self.assertEqual(df.coords["y"].value.size, 401)
        self.assertEqual(df.coords["y"].dim, 1)
        self.assertEqual(len(df.variables.keys()), 1)
        self.assertEqual(df.variables[0], df.variables["Density of States (Source)"])
        self.assertEqual(df.variables[0].name, "Density of States (Source)")
        self.assertEqual(df.variables[0].unit, "1/eV/nm")
        self.assertEqual(df.variables[0].value.shape, (50, 401))
        self.assertEqual(df.metadata, {})


class TestOutput(unittest.TestCase):

    def test_data(self):
        df = outputs.DataFile(
            folder_nnp / "bandedges_2d.fld", product="nextnano++"
        )
        self.assertEqual(len(df.data), len(df.coords) + len(df.variables))
        for key, value in df.coords.items():
            self.assertEqual(df.data[key], value)
        for key, value in df.variables.items():
            self.assertEqual(df.data[key], value)

    def test_access_by_index(self):
        df = outputs.DataFile(
            folder_nnp / "bandedges_2d.fld", product="nextnano++"
        )
        for key, value in df.coords.items():
            self.assertEqual(df[key], value)
        for key, value in df.variables.items():
            self.assertEqual(df[key], value)

    def test_for_loop(self):
        df = outputs.DataFile(
            folder_nnp / "bandedges_2d.fld", product="nextnano++"
        )
        for i, dfi in enumerate(df):
            self.assertEqual(df.data[i], dfi)


class TestDataFolder(unittest.TestCase):
    def test_init(self):
        dummy_folder = Path("tests") / "dummy"
        self.assertRaises(ValueError, outputs.DataFolder, dummy_folder)

        test_folder = Path("tests") / "dfolder"
        datafolder = outputs.DataFolder(test_folder)
        self.assertTrue("another_folder" in datafolder.folders)
        self.assertTrue(
            str(test_folder / "file1") in datafolder.files
        )

        self.assertTrue(
            str(test_folder / "file42") in datafolder.files
        )
        self.assertFalse("NotExistedFile.py" in datafolder.files)
        self.assertFalse(
            str(test_folder / "NotExistedFile.py") in datafolder.files
        )

        

        self.assertTrue(datafolder.folders)
        self.assertTrue(datafolder.files)

        self.assertEqual(len(datafolder.files), 3)
        

    def test_navigation(self):
        tests_folder = "tests"
        datafolder = outputs.DataFolder(tests_folder)
        self.assertTrue("configs" in datafolder.__dict__)
        self.assertIsInstance(datafolder.configs, outputs.DataFolder)
        self.assertIsInstance(datafolder.datafiles.nextnano3, outputs.DataFolder)
        self.assertIsInstance(
            datafolder.datafiles.folders["nextnano++"], outputs.DataFolder
        )
        nnp_folder = datafolder.datafiles.folders["nextnano++"]
        self.assertTrue(Path(nnp_folder.fullpath).samefile(folder_nnp))
        self.assertTrue(
            str(folder_nnp / "bandedges_2d_old.fld") in nnp_folder.files
        )

    def test_filenames(self):

        datafolder = outputs.DataFolder(folder_nnp)
        self.assertIn("only_variables.in", datafolder.filenames())
        self.assertNotIn("only_variables_0.in", datafolder.filenames())
        self.assertNotIn(
            str(folder_nnp / "only_variables.in"), datafolder.filenames()
        )

        tests_folder = "tests"
        datafolder = outputs.DataFolder(tests_folder)

        self.assertIn("test_outputs.py", datafolder.filenames())

        self.assertNotIn("only_variables.in", datafolder.filenames())
        self.assertNotIn(
            str(folder_nnp / "only_variables.in"), datafolder.filenames()
        )

    def test_find(self):
        tests_folder = "tests"
        datafolder = outputs.DataFolder(tests_folder)
        self.assertEqual(len(datafolder.find("")), 11)

        self.assertIn(str(Path(tests_folder) / "__init__.py"), datafolder.find(""))
        self.assertIn(
            str(Path(tests_folder) / "test_commands.py"), datafolder.find("")
        )

        self.assertNotIn(
            str(folder_nnp / "only_variables.in"), datafolder.find("")
        )
        self.assertNotIn("only_variables.in", datafolder.find(""))

        self.assertNotIn(
            str(Path(tests_folder) / "__init__.py"), datafolder.find("test")
        )
        self.assertIn(
            str(Path(tests_folder) / "test_commands.py"), datafolder.find("test")
        )

        self.assertNotIn(
            str(folder_nnp / "only_variables.in"), datafolder.find("test")
        )
        self.assertNotIn("only_variables.in", datafolder.find("test"))
        self.assertNotIn(
            str(Path(tests_folder) / "test_commands.py"), datafolder.find("test_i")
        )
        self.assertEqual(len(datafolder.find(".vtr")), 0)

        datafolder = outputs.DataFolder(folder_nnp)
        self.assertNotIn(
            str(Path(tests_folder) / "__init__.py"), datafolder.find("tests")
        )
        self.assertNotIn(
            str(Path(tests_folder) / "test_commands.py"), datafolder.find("tests")
        )

        self.assertNotIn(
            str(folder_nnp / "only_variables.in"), datafolder.find("tests")
        )
        self.assertIn(
            str(folder_nnp / "only_variables.in"), datafolder.find("only")
        )
        self.assertEqual(len(datafolder.find("bandedges")), 7)
        self.assertNotIn("only_variables.in", datafolder.find("abls"))
        self.assertEqual(len(datafolder.find(".vtr")), 2)

        # deep
        datafolder = outputs.DataFolder(tests_folder)
        self.assertIn(
            str(Path(tests_folder) / "__init__.py"), datafolder.find("", deep=True)
        )
        self.assertIn(
            str(Path(tests_folder) / "test_commands.py"),
            datafolder.find("", deep=True),
        )
        self.assertIn(
            str(folder_nnp / "only_variables.in"),
            datafolder.find("", deep=True),
        )
        self.assertNotIn("only_variables.in", datafolder.find("", deep=True))

        self.assertNotIn(
            str(Path(tests_folder) / "__init__.py"),
            datafolder.find("test", deep=True),
        )
        self.assertIn(
            str(Path(tests_folder) / "test_commands.py"),
            datafolder.find("test", deep=True),
        )

        self.assertNotIn(
            str(folder_nnp / "only_variables.in"),
            datafolder.find("test", deep=True),
        )
        self.assertIn(
            str(folder_nnp / "only_variables.in"),
            datafolder.find("only", deep=True),
        )
        self.assertNotIn(
            str(folder_nnp / "only_variables.in"),
            datafolder.find("nextnano", deep=True),
        )
        self.assertNotIn("only_variables.in", datafolder.find("only"))
        self.assertNotIn(
            str(Path(tests_folder) / "test_commands.py"),
            datafolder.find("test_i", deep=True),
        )
        self.assertNotEqual(len(datafolder.find(".vtr", deep=True)), 0)

        datafolder = outputs.DataFolder(folder_nnp)

        self.assertNotIn(
            str(Path(tests_folder) / "__init__.py"), datafolder.find("", deep=True)
        )
        self.assertNotIn(
            str(Path(tests_folder) / "test_commands.py"),
            datafolder.find("", deep=True),
        )
        self.assertIn(
            str(folder_nnp / "only_variables.in"),
            datafolder.find("", deep=True),
        )
        self.assertNotIn("only_variables.in", datafolder.find("", deep=True))

        self.assertNotIn(
            str(folder_nnp / "only_variables.in"),
            datafolder.find("test", deep=True),
        )
        self.assertIn(
            str(folder_nnp / "only_variables.in"),
            datafolder.find("only", deep=True),
        )
        self.assertNotIn(
            str(folder_nnp / "only_variables.in"),
            datafolder.find("nextnano", deep=True),
        )
        self.assertNotIn("only_variables.in", datafolder.find("only"))
        self.assertNotIn(
            str(Path(tests_folder) / "test_commands.py"),
            datafolder.find("test_i", deep=True),
        )
        self.assertNotEqual(len(datafolder.find(".vtr", deep=True)), 0)
        self.assertEqual(len(datafolder.find("bandedges", deep=True)), 17)

    def test_find_multiple(self):
        tests_folder = "tests"
        datafolder = outputs.DataFolder(tests_folder)
        self.assertEqual(len(datafolder.find_multiple(("",))), 11)

        self.assertIn(
            str(Path(tests_folder) / "__init__.py"), datafolder.find_multiple(("",))
        )

        self.assertNotIn("only_variables.in", datafolder.find_multiple(("",)))

        datafolder = outputs.DataFolder(folder_nnp)
        self.assertNotIn(
            str(Path(tests_folder) / "__init__.py"),
            datafolder.find_multiple(("tests",)),
        )

        self.assertEqual(len(datafolder.find_multiple(("bandedges",))), 7)
        self.assertEqual(len(datafolder.find_multiple(("bandedges", "2d"))), 5)

        self.assertNotIn("only_variables.in", datafolder.find_multiple(("abls",)))

        self.assertEqual(len(datafolder.find_multiple((".vtr",))), 2)
        self.assertEqual(len(datafolder.find_multiple((".vt", "r"))), 2)

    def test_go_to(self):
        tests_folder = "tests"
        datafolder = outputs.DataFolder(tests_folder)
        self.assertIsInstance(datafolder.go_to("datafiles"), outputs.DataFolder)
        self.assertIsInstance(datafolder.go_to("__init__.py"), str)
        self.assertTrue(
            Path(datafolder.go_to("__init__.py")).samefile(
                Path(tests_folder) / "__init__.py"
            )
        )
        self.assertFalse(
            Path(datafolder.go_to("__init__.py")).samefile(Path(tests_folder))
        )

        datafolder_nnp = datafolder.go_to("datafiles", "nextnano++")

        self.assertTrue(Path(datafolder_nnp.fullpath).samefile(folder_nnp))
        

    def test_file(self):
        tests_folder = "tests"
        datafolder = outputs.DataFolder(tests_folder)
        self.assertRaises(ValueError, datafolder.file, "hello_world")

        self.assertEqual(
            datafolder.datafiles.folders["nextnano++"].file("bandedges.vtr"),
            str(folder_nnp / "bandedges.vtr"),
        )

        self.assertRaises(
            ValueError,
            datafolder.datafiles.folders["nextnano++"].file,
            "bandedgesxxxx.vtr",
        )
        warnings.filterwarnings("ignore", "More than one file match")
        self.assertTrue(datafolder.datafiles.folders["nextnano++"].file("bandedges"))
        warnings.filterwarnings("default")

    def test_make_tree(self):
        tests_folder = "tests"
        datafolder = outputs.DataFolder(tests_folder)

        tree_list = datafolder.make_tree()
        print(tree_list)
        self.assertIn("    configs/", tree_list)
        self.assertIn("        .nnconfig", tree_list)
        self.assertIn("        nextnano++/", tree_list)

        self.assertLess(tree_list.index("    configs/"), tree_list.index("        nextnano++/"))
        self.assertLess(tree_list.index("        .nnconfig"), tree_list.index("        nextnano++/"))
        self.assertLess(tree_list.index("    datafiles/"), tree_list.index("        nextnano++/"))

if __name__ == "__main__":
    unittest.main()



import os
from nextnanopy.utils.mycollections import DictList
from nextnanopy.inputs import InputFileTemplate
from nextnanopy.nnp.defaults import (
    parse_nnp_variable,
    is_nnp_variable,
    InputVariable_nnp,
)
from nextnanopy.negf.defaults import is_negf_input_text
from nextnanopy.utils.misc import savetxt


class InputFile(InputFileTemplate):
    def load_variables(self):
        variables = DictList()
        for i, line in enumerate(self.raw_lines):
            if not is_nnp_variable(line):
                continue
            name, value, comment = parse_nnp_variable(line)
            var = InputVariable_nnp(
                name=name, value=value, comment=comment, metadata={"line_idx": i}
            )
            variables[var.name] = var
        self.variables = variables
        return self.variables

    def save(self, fullpath=None, overwrite=False, automkdir=True, content = False, temp=False):
        """
        content=True invokes the demo feature of saving self.content instead of self.test

        be aware that the content=True ignores comments
        """
        if fullpath is None:
            if temp:
                folder = self._get_temp_dir()
                fullpath = os.path.join(folder, self.filename)
            elif self.fullpath is None:
                raise ValueError('Please, specify a fullpath')
            else:
                fullpath = self.fullpath
        if content:
            text = self.content.__str__()
        else:
            text = self.text
        self.fullpath = savetxt(fullpath=fullpath, text=text, overwrite=overwrite, automkdir=automkdir)
        return self.fullpath


    def validate(self):
        if not is_negf_input_text(self.raw_text):
            raise ValueError(f"Not a valid nextnano.NEGF++ input file")



import unittest
from pathlib import Path
from nextnanopy import commands
from nextnanopy.utils.formatting import _path, _bool

folder_nnp = Path("tests") / "datafiles" / "nextnano++"
folder_nn3 = Path("tests") / "datafiles" / "nextnano3"
folder_negf = Path("tests") / "datafiles" / "nextnano.NEGF"
folder_msb = Path("tests") / "datafiles" / "nextnano.MSB"


class TestCommands(unittest.TestCase):

    def test_commands_nnp(self):
        self.maxDiff = None
        inputfile = folder_nnp / "example.in"
        exe = Path("nextnano++") / "bin 64bit" / "nextnano++_Intel_64bit.exe"
        runmode = "--resume"
        no_file_options = "--autosave --logfile"
        license = Path("nextnanopy") / "License" / "License_nnp.lic"
        database = Path("nextnano++") / "Syntax" / "database_nnp.in"
        outputdirectory = Path("tests") / "datafiles"
        threads = 4
        cmd = f'"{exe}" {runmode} --license "{license}" --database "{database}" --threads {threads} --outputdirectory "{outputdirectory}" --noautooutdir {no_file_options} "{inputfile}"'
        kwargs = dict(
            inputfile=inputfile,
            runmode=runmode,
            exe=exe,
            license=license,
            database=database,
            outputdirectory=outputdirectory,
            threads=threads,
            no_file_options=no_file_options,
        )
        from nextnanopy.nnp.defaults import command_nnp

        self.assertEqual(command_nnp(**kwargs), cmd)
        self.assertEqual(commands.command(**kwargs), cmd)

    def test_commands_nn3(self):
        self.maxDiff = None
        inputfile = folder_nn3 / "example.in"
        exe = Path("nextnano++") / "bin 64bit" / "nextnano++_Intel_64bit.exe"
        license = Path("nextnanopy") / "License" / "License_nnp.lic"
        database = Path("nextnano++") / "Syntax" / "database_nnp.in"
        outputdirectory = Path("tests") / "datafiles"
        threads = 4
        debuglevel = 0
        cancel = -1
        softkill = -1
        no_file_options = "-log -parse"
        cmd = f'"{exe}" -license "{license}" -inputfile "{inputfile}" -database "{database}" -threads {threads} -outputdirectory "{outputdirectory}" -debuglevel {debuglevel} -cancel {cancel} -softkill {softkill} {no_file_options}'
        kwargs = dict(
            inputfile=inputfile,
            exe=exe,
            license=license,
            database=database,
            outputdirectory=outputdirectory,
            threads=threads,
            debuglevel=debuglevel,
            cancel=cancel,
            softkill=softkill,
            no_file_options=no_file_options,
        )
        from nextnanopy.nn3.defaults import command_nn3

        self.assertEqual(command_nn3(**kwargs), cmd)
        self.assertEqual(commands.command(**kwargs), cmd)

    def test_commands_negf_classic(self):
        self.maxDiff = None
        inputfile = folder_negf / "example.xml"
        exe = Path("nextnano.NEGF") / "nextnano.NEGF.exe"
        license = Path("License") / "License_nnQCL.lic"
        database = Path("nextnano.NEGF") / "Material_Database.xml"
        outputdirectory = Path("tests") / "datafiles"
        threads = 4
        cmd = f'"{exe}" "{inputfile}" "{outputdirectory}" "{database}" "{license}" -threads {threads}'
        kwargs = dict(
            inputfile=inputfile,
            exe=exe,
            license=license,
            database=database,
            outputdirectory=outputdirectory,
            threads=threads,
        )
        from nextnanopy.negf.defaults import command_negf_classic

        self.assertEqual(command_negf_classic(**kwargs), cmd)
        self.assertEqual(commands.command(**kwargs), cmd)

    # TODO: add test_commands_negf

    def test_commands_msb(self):
        from nextnanopy.msb.defaults import command_msb

        self.maxDiff = None
        inputfile = folder_msb / "example.msb"
        exe = Path("nextnano.MSB") / "nextnano.MSB.exe"
        license = Path("License") / "License_nnMSB.lic"
        database = Path("nextnano.MSB") / "Materials.xml"
        outputdirectory = Path("tests") / "datafiles"
        threads = 1

        cmd = f'"{exe}" --license "{license}" --database "{database}" --threads {threads} --outputdirectory "{outputdirectory}" --noautooutdir "{inputfile}"'
        kwargs = dict(
            inputfile=inputfile,
            exe=exe,
            license=license,
            database=database,
            outputdirectory=outputdirectory,
            threads=threads,
        )
        self.assertEqual(command_msb(**kwargs), cmd)
        self.assertEqual(commands.command(**kwargs), cmd)

    def test_path(self):
        self.assertEqual(_path("aa\nb.test"), '"aa\nb.test"')
        self.assertEqual(_path("aa\nb"), '"aa\nb"')
        self.assertEqual(_path(""), "")
        self.assertEqual(_path(2), '"2"')
        self.assertEqual(_path(None), None)

    def test_execute(self):
        self.assertRaises(
            ValueError,
            commands.execute,
            inputfile="",
            exe="",
            license="",
            database="",
            outputdirectory="",
        )
        self.assertRaises(
            ValueError,
            commands.execute,
            inputfile=Path("test") / "datafiles",
            exe="",
            license="",
            database="",
            outputdirectory="",
        )

    def test_bool(self):
        self.assertEqual(_bool(""), False)
        self.assertEqual(_bool(None), False)
        self.assertEqual(_bool("1"), True)
        self.assertEqual(_bool("0"), True)
        self.assertEqual(_bool(0), True)
        self.assertEqual(_bool(1), True)


if __name__ == "__main__":
    unittest.main()



from pathlib import Path
from nextnanopy.defaults import NNConfig
import unittest
import warnings


default_config_path = Path.home() / ".nextnanopy-config"


class Test_NNConfig(unittest.TestCase):
    def test_default_nn3(self):
        config = NNConfig()

        self.assertEqual(Path(config.fullpath), default_config_path)
        self.assertTrue("nextnano3" in config.validators.keys())
        self.assertTrue("nextnano3" in config.defaults.keys())
        self.assertTrue("nextnano3" in config.config.keys())
        options = [
            "exe",
            "license",
            "database",
            "threads",
            "outputdirectory",
            "debuglevel",
            "cancel",
            "softkill",
        ]
        for option in options:
            self.assertTrue(option in config.validators["nextnano3"].keys())
            self.assertTrue(option in config.defaults["nextnano3"].keys())
            self.assertTrue(option in config.config["nextnano3"].keys())
        self.assertTrue(Path(config.fullpath).is_file())

    def test_default_nnp(self):
        config = NNConfig()
        self.assertEqual(Path(config.fullpath), default_config_path)
        self.assertTrue("nextnano++" in config.validators.keys())
        self.assertTrue("nextnano++" in config.defaults.keys())
        self.assertTrue("nextnano++" in config.config.keys())
        options = ["exe", "license", "database", "threads", "outputdirectory"]
        for option in options:
            self.assertTrue(option in config.validators["nextnano++"].keys())
            self.assertTrue(option in config.defaults["nextnano++"].keys())
            self.assertTrue(option in config.config["nextnano++"].keys())
        self.assertTrue(Path(config.fullpath).is_file())

    def test_default_negf(self):
        config = NNConfig()
        self.assertEqual(Path(config.fullpath), default_config_path)
        self.assertTrue("nextnano.NEGF" in config.validators.keys())
        self.assertTrue("nextnano.NEGF" in config.defaults.keys())
        self.assertTrue("nextnano.NEGF" in config.config.keys())
        options = ["exe", "license", "database", "threads", "outputdirectory"]
        for option in options:
            self.assertTrue(option in config.validators["nextnano.NEGF"].keys())
            self.assertTrue(option in config.defaults["nextnano.NEGF"].keys())
            self.assertTrue(option in config.config["nextnano.NEGF"].keys())
        self.assertTrue(Path(config.fullpath).is_file())

    def test_default_nnevo(self):
        config = NNConfig()
        self.assertEqual(Path(config.fullpath), default_config_path)
        self.assertTrue("nextnanoevo" in config.validators.keys())
        self.assertTrue("nextnanoevo" in config.defaults.keys())
        self.assertTrue("nextnanoevo" in config.config.keys())
        options = ["license"]
        for option in options:
            self.assertTrue(option in config.validators["nextnanoevo"].keys())
            self.assertTrue(option in config.defaults["nextnanoevo"].keys())
            self.assertTrue(option in config.config["nextnanoevo"].keys())
        option = "exe"  # no exe for nnevo
        self.assertFalse(option in config.validators["nextnanoevo"].keys())
        self.assertFalse(option in config.defaults["nextnanoevo"].keys())
        self.assertFalse(option in config.config["nextnanoevo"].keys())
        self.assertTrue(Path(config.fullpath).is_file())

    def test_load_nn3(self):
        fullpath = Path("tests") / ".nextnanopy-config"
        config = NNConfig(fullpath)

        self.assertEqual(Path(config.fullpath), fullpath)
        self.assertTrue("nextnano3" in config.validators.keys())
        self.assertTrue("nextnano3" in config.defaults.keys())
        self.assertTrue("nextnano3" in config.config.keys())
        options = [
            "exe",
            "license",
            "database",
            "threads",
            "outputdirectory",
            "debuglevel",
            "cancel",
            "softkill",
        ]
        for option in options:
            self.assertTrue(option in config.validators["nextnano3"].keys())
            self.assertTrue(option in config.defaults["nextnano3"].keys())
            self.assertTrue(option in config.config["nextnano3"].keys())
        self.assertTrue(Path(config.fullpath).is_file())
        if Path(config.fullpath).is_file():
            Path(config.fullpath).unlink()

        self.assertEqual(config.config["nextnano3"]["exe"], "")
        config.set("nextnano3", "exe", "some_path")
        self.assertEqual(config.config["nextnano3"]["exe"], "some_path")
        config.to_default()
        self.assertEqual(config.config["nextnano3"]["exe"], "")
        if Path(config.fullpath).is_file():
            Path(config.fullpath).unlink()
        fullpath_new = Path("tests") / "test.nnconfig"
        self.assertFalse(fullpath_new.is_file())
        config.save(fullpath_new)
        self.assertTrue(fullpath_new.is_file())
        self.assertEqual(Path(config.fullpath), fullpath_new)
        if Path(config.fullpath).is_file():
            Path(config.fullpath).unlink()

    def test_load_nnp(self):
        fullpath = Path("tests") / ".nextnanopy-config"
        config = NNConfig(fullpath)

        self.assertEqual(Path(config.fullpath), fullpath)
        self.assertTrue("nextnano++" in config.validators.keys())
        self.assertTrue("nextnano++" in config.defaults.keys())
        self.assertTrue("nextnano++" in config.config.keys())
        options = ["exe", "license", "database", "threads", "outputdirectory"]
        for option in options:
            self.assertTrue(option in config.validators["nextnano++"].keys())
            self.assertTrue(option in config.defaults["nextnano++"].keys())
            self.assertTrue(option in config.config["nextnano++"].keys())
        self.assertTrue(Path(config.fullpath).is_file())
        if Path(config.fullpath).is_file():
            Path(config.fullpath).unlink()

        self.assertEqual(config.config["nextnano++"]["exe"], "")
        config.set("nextnano++", "exe", "some_path")
        self.assertEqual(config.config["nextnano++"]["exe"], "some_path")
        config.to_default()
        self.assertEqual(config.config["nextnano++"]["exe"], "")
        if Path(config.fullpath).is_file():
            Path(config.fullpath).unlink()
        fullpath_new = Path("tests") / "test.nnconfig"
        self.assertFalse(fullpath_new.is_file())
        config.save(fullpath_new)
        self.assertTrue(fullpath_new.is_file())
        self.assertEqual(Path(config.fullpath), fullpath_new)
        if Path(config.fullpath).is_file():
            Path(config.fullpath).unlink()

    def test_load_negf(self):
        fullpath = Path("tests") / ".nextnanopy-config"
        config = NNConfig(fullpath)

        self.assertEqual(Path(config.fullpath), fullpath)
        self.assertTrue("nextnano.NEGF" in config.validators.keys())
        self.assertTrue("nextnano.NEGF" in config.defaults.keys())
        self.assertTrue("nextnano.NEGF" in config.config.keys())
        options = ["exe", "license", "database", "threads", "outputdirectory"]
        for option in options:
            self.assertTrue(option in config.validators["nextnano.NEGF"].keys())
            self.assertTrue(option in config.defaults["nextnano.NEGF"].keys())
            self.assertTrue(option in config.config["nextnano.NEGF"].keys())
        self.assertTrue(Path(config.fullpath).is_file())
        if Path(config.fullpath).is_file():
            Path(config.fullpath).unlink()

        self.assertEqual(config.config["nextnano.NEGF"]["exe"], "")
        config.set("nextnano.NEGF", "exe", "some_path")
        self.assertEqual(config.config["nextnano.NEGF"]["exe"], "some_path")
        config.to_default()
        self.assertEqual(config.config["nextnano.NEGF"]["exe"], "")
        if Path(config.fullpath).is_file():
            Path(config.fullpath).unlink()
        fullpath_new = Path("tests") / "test.nnconfig"
        self.assertFalse(fullpath_new.is_file())
        config.save(fullpath_new)
        self.assertTrue(fullpath_new.is_file())
        self.assertEqual(Path(config.fullpath), fullpath_new)
        if Path(config.fullpath).is_file():
            Path(config.fullpath).unlink()

    def test_get_unsupported_products(self):
        filepath = Path("tests") / "configs" / ".nnconfig_unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            config = NNConfig(filepath)
        unsupported_products = set(config.get_unsupported_products())
        self.assertEqual(
            unsupported_products,
            {"nextnano.NEGF++", "nextnano_nonexistent"},
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
from nextnanopy.utils.mycollections import DictList
from nextnanopy.outputs import Output, AvsAscii, Vtk, DataFileTemplate, Dat
from nextnanopy.utils.datasets import Variable, Coord
import re


class DataFile(DataFileTemplate):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath, product="nextnano.MSB")
        self.load()

    def get_loader(self):
        if self.extension in [".v", ".fld", ".coord"]:
            loader = AvsAscii
        elif self.extension == ".vtr":
            loader = Vtk
        elif self.extension == ".txt":
            raise NotImplementedError(
                f"Loading nextnano.MSB datafiles with extension.txt is not implemented yet"
            )
        elif self.extension == ".dat":
            loader = Dat
        else:
            raise NotImplementedError(
                f"Loading datafile with extension {self.extension} is not implemented yet"
            )
        return loader



import unittest
from pathlib import Path
from nextnanopy.utils import mycollections
from nextnanopy import defaults
from nextnanopy.nnp import defaults as nnp_defaults
from nextnanopy.nn3 import defaults as nn3_defaults
from nextnanopy.negf import defaults as negf_defaults
from nextnanopy.msb import defaults as msb_defaults


folder_nnp = Path("tests") / "datafiles" / "nextnano++"
folder_nn3 = Path("tests") / "datafiles" / "nextnano3"
folder_negf = Path("tests") / "datafiles" / "nextnano.NEGF"


class TestDictlist(unittest.TestCase):
    def test_dictlist(self):
        a = mycollections.DictList(a=2, b=3, c=4)
        self.assertEqual(a[0], 2)
        self.assertEqual(a["a"], 2)
        self.assertEqual(a[1], 3)
        self.assertEqual(a["b"], 3)
        self.assertEqual(list(a.keys()), ["a", "b", "c"])
        self.assertEqual(list(a.values()), [2, 3, 4])


class TestFormatting(unittest.TestCase):
    def test_nn3(self):
        fullpath = folder_nn3 / "example.in"
        self.assertTrue(nn3_defaults.is_nn3_input_file(fullpath))
        self.assertFalse(nnp_defaults.is_nnp_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negf_classic_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negf_input_file(fullpath))
        self.assertEqual(defaults.input_file_type(fullpath), "nextnano3")
        self.assertEqual(defaults.get_fmt("nextnano3")["var_char"], "%")
        self.assertEqual(defaults.get_fmt("nextnano3")["com_char"], "!")
        self.assertEqual(
            defaults.get_fmt("nextnano3")["input_pattern"], "$end_simulation-dimension"
        )

    def test_nnp(self):
        fullpath = folder_nnp / "example.in"
        self.assertFalse(nn3_defaults.is_nn3_input_file(fullpath))
        self.assertTrue(nnp_defaults.is_nnp_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negf_classic_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negf_input_file(fullpath))
        self.assertEqual(defaults.input_file_type(fullpath), "nextnano++")
        self.assertEqual(defaults.get_fmt("nextnano++")["var_char"], "$")
        self.assertEqual(defaults.get_fmt("nextnano++")["com_char"], "#")
        self.assertEqual(defaults.get_fmt("nextnano++")["input_pattern"], "global{")

    def test_negf_classic(self):
        fullpath = folder_negf / "example.xml"
        self.assertFalse(nn3_defaults.is_nn3_input_file(fullpath))
        self.assertFalse(nnp_defaults.is_nnp_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negf_input_file(fullpath))
        self.assertTrue(negf_defaults.is_negf_classic_input_file(fullpath))
        self.assertEqual(defaults.input_file_type(fullpath), "nextnano.NEGF_classic")
        self.assertEqual(defaults.get_fmt("nextnano.NEGF_classic")["var_char"], "$")
        self.assertEqual(defaults.get_fmt("nextnano.NEGF_classic")["com_char"], "<!--")
        self.assertEqual(
            defaults.get_fmt("nextnano.NEGF_classic")["input_pattern"], "<Simulation"
        )

    def test_negf(self):
        fullpath = folder_negf / "Minimal_InputFile.negf"
        self.assertFalse(nn3_defaults.is_nn3_input_file(fullpath))
        self.assertFalse(nnp_defaults.is_nnp_input_file(fullpath))
        self.assertFalse(negf_defaults.is_negf_classic_input_file(fullpath))
        self.assertTrue(negf_defaults.is_negf_input_file(fullpath))
        self.assertEqual(defaults.input_file_type(fullpath), "nextnano.NEGF")
        self.assertEqual(defaults.get_fmt("nextnano.NEGF")["var_char"], "$")
        self.assertEqual(defaults.get_fmt("nextnano.NEGF")["com_char"], "#")
        self.assertEqual(
            defaults.get_fmt("nextnano.NEGF")["input_pattern"], "nextnano.NEGF{"
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
from nextnanopy.utils.mycollections import DictList
from nextnanopy.outputs import Output, AvsAscii, Vtk, DataFileTemplate, Dat
from nextnanopy.nnp.defaults import (
    parse_nnp_variable,
    is_nnp_variable,
    InputVariable_nnp,
)
from nextnanopy.utils.datasets import Variable, Coord
from nextnanopy.utils.formatting import best_str_to_name_unit
from functools import partial


class DataFile(DataFileTemplate):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath, product="nextnano++")
        self.load(**loader_kwargs)

    def get_loader(self):
        if self.extension in [".v", ".fld", ".coord"]:
            loader = AvsAscii
        elif self.extension == ".vtr":
            loader = Vtk
        elif self.extension == ".txt":
            loader = self._find_txt_loader()
        elif self.extension == ".dat":
            loader = Dat
        else:
            raise NotImplementedError(
                f"Loading datafile with extension {self.extension} is not implemented yet"
            )
        return loader

    def _find_txt_loader(self):
        if self.filename_only in ["variables_input", "variables_database"]:
            loader = InputVariables
        elif self.filename_only == "materials":
            raise NotImplementedError(f"Loading materials.txt is not implemented yet")
        elif self.filename_only == "total_charges":
            loader = TotalCharges
        else:
            loader = partial(Dat, FirstVarIsCoordFlag=False)
        return loader


class InputVariables(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load()

    def load(self):
        self.load_raw()
        self.load_variables()

    def load_raw(self):
        with open(self.fullpath, "r") as f:
            self.raw_lines = f.readlines()

    def load_variables(self):
        variables = DictList()
        for i, line in enumerate(self.raw_lines):
            if not is_nnp_variable(line):
                continue
            name, value, comment = parse_nnp_variable(line)
            var = InputVariable_nnp(
                name=name, value=value, comment=comment, metadata={"line_idx": i}
            )
            variables[var.name] = var
        self.variables = variables
        return self.variables


class TotalCharges(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load()

    def load(self):
        self.load_raw()
        self.load_variables()

    def load_raw(self):
        with open(self.fullpath, "r") as f:
            self.raw_lines = f.readlines()

    def load_variables(self):
        variables = DictList()
        for i, line in enumerate(self.raw_lines):
            try:
                name, value, unit = line.split()
            except Exception as e:
                raise ValueError(
                    f"An error occurred while loading total_charges file:{e}"
                )

            try:
                value = float(value)
            except Exception as e:
                raise ValueError(
                    f"An error occurred while loading total_charges file:{e}"
                )

            # temp solution to omit ":" in the end of the line
            if name[-1] == ":":
                name = name[:-1]

            var = Variable(name=name, unit=unit, value=value)
            variables[var.name] = var

        self.variables = variables
        return self.variables


# class Dat(Output):
#     def __init__(self, fullpath, FirstVarIsCoordFlag=True):
#         super().__init__(fullpath)
#         self.load(FirstVarIsCoordFlag)

#     def load(self, FirstVarIsCoordFlag):
#         self.load_metadata(FirstVarIsCoordFlag)
#         self.load_data()

#     def _get_headers(self):
#         headers = []
#         with open(self.fullpath, "r") as f:
#             for line in f:
#                 try:
#                     float(line.split()[0])
#                     break
#                 except:
#                     headers.append(line)
#         return headers

#     def load_metadata(self, FirstVarIsCoordFlag):
#         metadata = {}
#         headers = self._get_headers()
#         metadata["headers"] = headers
#         metadata["skip_rows"] = len(headers)

#         if len(headers) == 0:
#             raise NotImplementedError(".dat file without header")
#         else:
#             header = headers[-1]  # take the last one by default
#         header = header.split()
#         ndim = 0
#         dkeys = []

#         # FirstVarIsCoordFlag = True
#         for i, column in enumerate(header):
#             key, unit = best_str_to_name_unit(column, default_unit=None)
#             metadata[i] = {"name": key, "unit": unit}
#             # if key.lower() in ['x', 'y', 'z', 'position']:
#             #    ndim += 1
#             #    dkeys.append(i)
#             if FirstVarIsCoordFlag:
#                 ndim += 1
#                 dkeys.append(i)
#                 FirstVarIsCoordFlag = False  # use this to change behaviour of recognition of coords and variables
#         metadata["ndim"] = ndim
#         metadata["dkeys"] = dkeys
#         self.metadata.update(metadata)
#         return metadata

#     def load_data(self):
#         data = []
#         meta = self.metadata
#         # with open(self.fullpath, 'r') as f:
#         #     for i, line in enumerate(f):
#         #         if i < meta['skip_rows']:
#         #             continue
#         #         line = line.replace('\n', '').strip().split()
#         #         if line:
#         #             data.append(line)
#         data = np.loadtxt(self.fullpath, skiprows=meta["skip_rows"])
#         data = np.array(data, dtype=float).T  # columns 1st index
#         coords, variables = DictList(), DictList()
#         dims = []
#         for i, values in enumerate(data):
#             vm = meta[i]
#             if i in meta["dkeys"]:
#                 # values = np.unique(values)
#                 dims.append(values.size)
#                 var = Coord(name=vm["name"], unit=vm["unit"], dim=i, value=values)
#                 coords[var.name] = var
#             else:
#                 if dims:
#                     values = values.reshape(*dims)
#                 var = Variable(name=vm["name"], unit=vm["unit"], value=values)
#                 variables[var.name] = var
#         self.coords = coords
#         self.variables = variables
#         return coords, variables

#     def save(self, new_location, extension="dat"):
#         with open(new_location, "w") as f:
#             # Write headers
#             for header in self.metadata["headers"]:
#                 f.write(header)

#             # Write data
#             combined_data = [coord.value for coord in self.coords] + [
#                 variable.value for variable in self.variables
#             ]

#             data = np.column_stack(combined_data).transpose()
#             np.savetxt(f, data.T)

#         # Optionally update the fullpath attribute to the new location
#         self.fullpath = new_location



from pathlib import Path
import warnings
from nextnanopy.utils.config import Config

products = [
    "nextnano++",
    "nextnano3",
    "nextnano.NEGF",
    "nextnano.NEGF_classic",
    "nextnano.MSB",
    "nextnanoevo",
]
config_default_path = Path.home() / ".nextnanopy-config"
messages = {
    "load_input": [None, None],
    "save_input": [None, None],
    "execute_input": [None, None],
    "load_output": [None, None],
}


def get_InputFile(product):
    if product == "nextnano++":
        from nextnanopy.nnp.inputs import InputFile
    elif product == "nextnano3":
        from nextnanopy.nn3.inputs import InputFile
    elif product == "nextnano.NEGF_classic":
        from nextnanopy.negf.inputs_classic import InputFile
    elif product == "nextnano.NEGF":
        from nextnanopy.negf.inputs import InputFile
    elif product == "nextnano.MSB":
        from nextnanopy.msb.inputs import InputFile
    elif product == "nextnanoevo":
        raise ValueError("There is no InputFile format for nextnanoevo")
    elif product == "not valid":
        from nextnanopy.inputs import InputFileTemplate as InputFile
    else:
        raise ValueError(f"{product} is not valid")
    return InputFile


def get_DataFile(product):
    if product == "nextnano3":
        from nextnanopy.nn3.outputs import DataFile
    elif product == "nextnano++":
        from nextnanopy.nnp.outputs import DataFile
    elif product == "nextnano.NEGF_classic" or product == "nextnano.NEGF":
        from nextnanopy.negf.outputs import DataFile
    elif product == "nextnano.MSB":
        from nextnanopy.negf.outputs import DataFile
    elif product == "nextnanoevo":
        raise ValueError("There is no DataFile format for nextnanoevo")
    else:
        raise ValueError(f"{product} is not valid")
    return DataFile


def get_command(product):
    if product == "nextnano++":
        from nextnanopy.nnp.defaults import command_nnp as command
    elif product == "nextnano3":
        from nextnanopy.nn3.defaults import command_nn3 as command
    elif product == "nextnano.NEGF_classic":
        from nextnanopy.negf.defaults import command_negf_classic as command
    elif product == "nextnano.NEGF":
        from nextnanopy.negf.defaults import command_negf as command
    elif product == "nextnano.MSB":
        from nextnanopy.msb.defaults import command_msb as command
    elif product == "nextnanoevo":
        raise ValueError("There is no command format for nextnanoevo")
    else:
        raise ValueError(f"{product} is not valid")
    return command


def get_fmt(product):
    if product == "nextnano++":
        from nextnanopy.nnp.defaults import fmt
    elif product == "nextnano3":
        from nextnanopy.nn3.defaults import fmt
    elif product == "nextnano.NEGF_classic":
        from nextnanopy.negf.defaults import fmt_classic as fmt
    elif product == "nextnano.NEGF":
        from nextnanopy.negf.defaults import fmt
    elif product == "nextnano.MSB":
        from nextnanopy.msb.defaults import fmt
    elif product == "nextnanoevo":
        raise ValueError("There is no formatting defaults for nextnanoevo")
    else:
        raise ValueError(f"{product} is not valid")
    return fmt


def input_file_type(fullpath):
    from nextnanopy.nnp.defaults import is_nnp_input_file
    from nextnanopy.nn3.defaults import is_nn3_input_file
    from nextnanopy.negf.defaults import (
        is_negf_classic_input_file,
        is_negf_input_file,
    )
    from nextnanopy.msb.defaults import is_msb_input_file

    if is_nn3_input_file(fullpath):
        return "nextnano3"
    elif is_nnp_input_file(fullpath):
        return "nextnano++"
    elif is_negf_classic_input_file(fullpath):
        return "nextnano.NEGF_classic"
    elif is_negf_input_file(fullpath):
        return "nextnano.NEGF"
    elif is_msb_input_file(fullpath):
        return "nextnano.MSB"
    else:
        return "not valid"


def input_text_type(text):
    from nextnanopy.nnp.defaults import is_nnp_input_text
    from nextnanopy.nn3.defaults import is_nn3_input_text
    from nextnanopy.negf.defaults import (
        is_negf_classic_input_text,
        is_negf_input_text,
    )
    from nextnanopy.msb.defaults import is_msb_input_text

    if is_nn3_input_text(text):
        return "nextnano3"
    elif is_nnp_input_text(text):
        return "nextnano++"
    elif is_negf_classic_input_text(text):
        return "nextnano.NEGF_classic"
    elif is_negf_input_text(text):
        return "nextnano.NEGF"
    elif is_msb_input_text(text):
        return "nextnano.MSB"
    else:
        return "not valid"


def get_config_validators():
    config_validator = {product: _get_config_validator(product) for product in products}
    return config_validator


def get_config_defaults():
    config_defaults = {product: _get_config_default(product) for product in products}
    return config_defaults


def _get_config_validator(product):
    if product == "nextnano++":
        from nextnanopy.nnp.defaults import config_validator
    elif product == "nextnano3":
        from nextnanopy.nn3.defaults import config_validator
    elif product == "nextnano.NEGF_classic" or product == "nextnano.NEGF":
        from nextnanopy.negf.defaults import config_validator
    elif product == "nextnano.MSB":
        from nextnanopy.msb.defaults import config_validator
    elif product == "nextnanoevo":
        from nextnanopy.nnevo.defaults import config_validator
    else:
        raise ValueError(f"{product} is not valid")
    return config_validator


def _get_config_default(product):
    if product == "nextnano++":
        from nextnanopy.nnp.defaults import config_default
    elif product == "nextnano3":
        from nextnanopy.nn3.defaults import config_default
    elif product == "nextnano.NEGF_classic" or product == "nextnano.NEGF":
        from nextnanopy.negf.defaults import config_default
    elif product == "nextnano.MSB":
        from nextnanopy.msb.defaults import config_default
    elif product == "nextnanoevo":
        from nextnanopy.nnevo.defaults import config_default
    else:
        raise ValueError(f"{product} is not valid")
    return config_default


class NNConfig(Config):
    def __init__(self, fullpath=None):
        self.default_fullpath = config_default_path
        validators = get_config_validators()
        self.defaults = get_config_defaults()
        if fullpath is None:
            fullpath = self.default_fullpath
        super().__init__(fullpath, validators)
        if not Path(fullpath).is_file():
            self.reset()
        elif not self._complete:
            # TODO add tests for these behaviour
            # ensures smooth transition to additional product: NEGF++
            self.update_with_defaults()
            self.save()

        unsupported_products = self.get_unsupported_products()
        if unsupported_products:
            warnings.warn(
                f"Unsupported products in config file: {unsupported_products} will be ignored. To not see this message, please remove unsupported products from the config file: {self.fullpath}"
                "Note: nextnano.NEGF++ was renamed to nextnano.NEGF, nextnano.NEGF was renamed to nextnano.NEGF_classic. Please check the documentation for more details.",
            )

    def to_default(self):
        for section in self.defaults.keys():
            if section not in self.sections:
                self.add_section(section)
            for option, value in self.defaults[section].items():
                self.set(section, option, value)

    def reset(self):
        self.to_default()
        self.save()

    @property
    def _complete(self):
        return self.check_complete()

    def check_complete(self):
        for section in self.defaults.keys():
            if section not in self.sections:
                return False
        return True

    def get_unsupported_products(self):
        unsupported_products = []
        for section in self.sections:
            if section not in self.defaults.keys():
                unsupported_products.append(section)
        return unsupported_products

    def update_with_defaults(self):
        for section in self.defaults.keys():
            if section not in self.sections:
                self.add_section(section)
                for option, value in self.defaults[section].items():
                    self.set(section, option, value)



import unittest
import os
import tempfile
from pathlib import Path
from nextnanopy.inputs import InputFile, Sweep


def delete_files(start, directory=Path.cwd(), exceptions=None):
    for fpath in Path(directory).iterdir():
        if fpath.name.startswith(start):
            if exceptions:
                if fpath.name in exceptions:
                    continue
                else:
                    fpath.unlink()
            else:
                fpath.unlink()


folder_nnp = Path("tests") / "datafiles" / "nextnano++"
folder_nn3 = Path("tests") / "datafiles" / "nextnano3"
folder_negf = Path("tests") / "datafiles" / "nextnano.NEGF"
folder_msb = Path("tests") / "datafiles" / "nextnano.MSB"

class Test_nnp(unittest.TestCase):
    def test_load(self):
        fullpath = folder_nnp / "only_variables.in"

        file = InputFile(fullpath)
        self.assertEqual(file.product, "nextnano++")

        self.assertEqual(len(file.variables.keys()), 7)
        self.assertEqual(file.variables["float"].name, "float")
        self.assertEqual(file.variables["float"].value, float(0.0))
        self.assertEqual(file.variables["float"].comment, "float number (a.u.)")

        self.assertEqual(file.variables["int"].name, "int")
        self.assertEqual(file.variables["int"].value, int(0))
        self.assertEqual(file.variables["int"].comment, "int number (random = 3)")

        self.assertEqual(file.variables["str"].name, "str")
        self.assertEqual(file.variables["str"].value, '"test"')
        self.assertEqual(
            file.variables["str"].comment, "first comments # second comment"
        )

        self.assertEqual(file.variables["reference"].name, "reference")
        self.assertEqual(file.variables["reference"].value, "$int - 1")

        self.assertEqual(file.variables["NoSpacing"].name, "NoSpacing")
        self.assertEqual(file.variables["NoSpacing"].value, int("7"))
        self.assertEqual(file.variables["NoSpacing"].comment, "this is a comment")
        self.assertEqual(
            file.variables["NoSpacing"].text, "$NoSpacing = 7 # this is a comment"
        )

        self.assertEqual(file.variables["MAYUS"].name, "MAYUS")
        self.assertEqual(file.variables["MAYUS"].value, "TEXT")
        self.assertEqual(file.variables["MAYUS"].comment, "")
        self.assertEqual(file.variables["MAYUS"].text, "$MAYUS = TEXT")

        fullpath = folder_nnp / "virtual_file.in"
        self.assertRaises(FileNotFoundError, InputFile, fullpath)

    def test_get_variables(self):
        fullpath = folder_nn3 / "only_variables.in"
        file = InputFile(fullpath)

        self.assertEqual(file.variables["float"], file.get_variable("float"))
        self.assertRaises(KeyError, file.get_variable, name="new_variable")

    def test_set_variables(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)
        file.set_variable(
            "float", 1e-5, "some comment", "some unit"
        )  # unit is not visible in nn++

        self.assertEqual(file.variables["float"].name, "float")
        self.assertEqual(file.variables["float"].value, 1e-5)
        self.assertEqual(file.variables["float"].comment, "some comment")
        self.assertEqual(
            file.variables["float"].text, f"$float = {str(1e-5)} # some comment"
        )
        self.assertEqual(file.lines[3], f"$float = {str(1e-5)} # some comment")

        file.set_variable("float", value=1e-7)
        self.assertEqual(file.variables["float"].name, "float")
        self.assertEqual(file.variables["float"].value, 1e-7)
        self.assertEqual(file.variables["float"].comment, "some comment")
        self.assertEqual(
            file.variables["float"].text, f"$float = {str(1e-7)} # some comment"
        )
        self.assertEqual(file.lines[3], f"$float = {str(1e-7)} # some comment")

        file.set_variable("float", comment="new comment")
        self.assertEqual(file.variables["float"].name, "float")
        self.assertEqual(file.variables["float"].value, 1e-7)
        self.assertEqual(file.variables["float"].comment, "new comment")
        self.assertEqual(
            file.variables["float"].text, f"$float = {str(1e-7)} # new comment"
        )
        self.assertEqual(file.lines[3], f"$float = {str(1e-7)} # new comment")

        file.set_variable("float", value=0)
        self.assertEqual(file.variables["float"].value, 0)
        file.set_variable("float", value="0")
        self.assertEqual(file.variables["float"].value, "0")
        self.assertRaises(KeyError, file.set_variable, name="new_variable")
        self.assertRaises(KeyError, file.set_variable, name="new_variable")

    def test_fullpath(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)

        self.assertEqual(Path(file.fullpath), fullpath)
        self.assertEqual(
            Path(file.save(file.fullpath, overwrite=False)),
            folder_nnp / "only_variables_0.in",
        )
        Path(file.fullpath).unlink()

    def test_config(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)
        from nextnanopy import config

        for key, value in config.config["nextnano++"].items():
            self.assertEqual(file.default_command_args[key], value)

    def test_text(self):
        fullpath_onlyvar = folder_nnp / "only_variables.in"
        fullpath_example = folder_nnp / "example.in"
        file = InputFile(fullpath_onlyvar)
        text = file.text

        new_file = InputFile(fullpath_example)
        new_file.text = text
        self.assertEqual(new_file.product, "nextnano++")
        self.assertEqual(new_file.text, text)
        self.assertEqual(new_file.lines, file.lines)
        self.assertEqual(new_file.variables["MAYUS"].name, "MAYUS")
        self.assertEqual(new_file.variables["MAYUS"].value, "TEXT")
        self.assertEqual(new_file.variables["MAYUS"].comment, "")
        self.assertEqual(new_file.variables["MAYUS"].text, "$MAYUS = TEXT")
        self.assertEqual(
            Path(new_file.save(file.fullpath, overwrite=False)),
            folder_nnp / "only_variables_0.in",
        )
        Path(new_file.fullpath).unlink()

    def test_set_and_save(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)
        file.set_variable(name="float", value=0.4)
        self.assertAlmostEqual(file.variables["float"].value, 0.4)

        self.addCleanup((folder_nnp / "only_variables_0.in").unlink)
        file.save()
        self.assertTrue((folder_nnp / "only_variables_0.in").is_file())

    def test_save_temp(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)
        file.save(temp=True)
        self.assertTrue(Path(file.fullpath).is_file())
        try:
            self.assertTrue(Path(file.fullpath).is_file())
            # Check that the file is in the system temp directory
            temp_dir = Path(tempfile.gettempdir()).resolve()
            file_dir = Path(file.fullpath).resolve().parent
            self.assertTrue(str(file_dir).startswith(str(temp_dir)))
        finally:
            if Path(file.fullpath).exists():
                Path(file.fullpath).unlink()

    def test_same_dir_saving(self):
        current_directory = Path.cwd()
        self.addCleanup(os.chdir, current_directory)

        os.chdir(folder_nnp)

        path = "only_variables.in"
        file = InputFile(path)
        file.set_variable(name="float", value=0.3333)
        self.assertAlmostEqual(file.variables["float"].value, 0.3333)
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=".",
            exceptions=["only_variables.in"],
        )
        file.save()
        self.assertTrue(Path("only_variables_0.in").is_file())

    ###content tests

    def test_content_get(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)

        self.assertIsNotNone(file.content)
        self.assertEqual(file.content[0], "$float = 0.0 ")
        self.assertEqual(file.content[-1].name, "global")

    def test_content_set(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)

        file.content[0] = "$DUMMY = 1"
        self.assertEqual(file.content[0], "$DUMMY = 1")
        file.content["_entry_0"] = "DUMMY LINE"
        self.assertEqual(file.content[0], "DUMMY LINE")


class Test_nn3(unittest.TestCase):
    def test_load(self):
        fullpath = folder_nn3 / "only_variables.in"

        file = InputFile(fullpath)
        self.assertEqual(file.product, "nextnano3")

        self.assertEqual(len(file.variables.keys()), 7)
        self.assertEqual(file.variables["float"].name, "float")
        self.assertEqual(file.variables["float"].value, float(0.0))
        self.assertEqual(file.variables["float"].comment, "float number (a.u.)")

        self.assertEqual(file.variables["int"].name, "int")
        self.assertEqual(file.variables["int"].value, int(0))
        self.assertEqual(file.variables["int"].comment, "int number (random = 3)")

        self.assertEqual(file.variables["str"].name, "str")
        self.assertEqual(file.variables["str"].value, '"test"')
        self.assertEqual(
            file.variables["str"].comment, "first comments # second comment"
        )

        self.assertEqual(file.variables["reference"].name, "reference")
        self.assertEqual(file.variables["reference"].value, "$int - 1")

        self.assertEqual(file.variables["NoSpacing"].name, "NoSpacing")
        self.assertEqual(file.variables["NoSpacing"].value, int("7"))
        self.assertEqual(file.variables["NoSpacing"].comment, "this is a comment")

        self.assertEqual(file.variables["MAYUS"].name, "MAYUS")
        self.assertEqual(file.variables["MAYUS"].value, "TEXT")
        self.assertEqual(file.variables["MAYUS"].comment, "")

        fullpath = folder_nn3 / "virtual_file.in"
        self.assertRaises(FileNotFoundError, InputFile, fullpath)

    def test_get_variables(self):
        fullpath = folder_nn3 / "only_variables.in"
        file = InputFile(fullpath)

        self.assertEqual(file.variables["float"], file.get_variable("float"))
        self.assertRaises(KeyError, file.get_variable, name="new_variable")

    def test_set_variables(self):
        fullpath = folder_nn3 / "only_variables.in"
        file = InputFile(fullpath)
        file.set_variable("float", 1e-5, "some comment")

        self.assertEqual(file.variables["float"].name, "float")
        self.assertEqual(file.variables["float"].value, 1e-5)
        self.assertEqual(file.variables["float"].comment, "some comment")
        self.assertEqual(
            file.variables["float"].text, f"%float = {str(1e-5)} ! some comment"
        )
        self.assertEqual(file.lines[3], f"%float = {str(1e-5)} ! some comment")

        file.set_variable("float", value=1e-7)
        self.assertEqual(file.variables["float"].name, "float")
        self.assertEqual(file.variables["float"].value, 1e-7)
        self.assertEqual(file.variables["float"].comment, "some comment")
        self.assertEqual(
            file.variables["float"].text, f"%float = {str(1e-7)} ! some comment"
        )
        self.assertEqual(file.lines[3], f"%float = {str(1e-7)} ! some comment")

        file.set_variable("float", comment="new comment")
        self.assertEqual(file.variables["float"].name, "float")
        self.assertEqual(file.variables["float"].value, 1e-7)
        self.assertEqual(file.variables["float"].comment, "new comment")
        self.assertEqual(
            file.variables["float"].text, f"%float = {str(1e-7)} ! new comment"
        )
        self.assertEqual(file.lines[3], f"%float = {str(1e-7)} ! new comment")

        file.set_variable("float", value=0)
        self.assertEqual(file.variables["float"].value, 0)
        file.set_variable("float", value="0")
        self.assertEqual(file.variables["float"].value, "0")
        self.assertRaises(KeyError, file.set_variable, name="new_variable")

    def test_fullpath(self):
        fullpath = folder_nn3 / "only_variables.in"
        file = InputFile(fullpath)

        self.assertEqual(Path(file.fullpath), fullpath)
        self.assertEqual(
            Path(file.save(file.fullpath, overwrite=False)),
            folder_nn3 / "only_variables_0.in",
        )
        Path(file.fullpath).unlink()

    def test_config(self):
        fullpath = folder_nn3 / "only_variables.in"
        file = InputFile(fullpath)
        from nextnanopy import config

        for key, value in config.config["nextnano3"].items():
            self.assertEqual(file.default_command_args[key], value)

    def test_save(self):
        fullpath = folder_nn3 / "only_variables.in"
        file = InputFile(fullpath)
        new_folder = folder_nn3 / "temp"
        new_file = new_folder / "example_copy.in"
        self.assertRaises(
            FileNotFoundError, file.save, new_file, overwrite=True, automkdir=False
        )
        self.assertEqual(Path(file.save(new_file, overwrite=True, automkdir=True)), new_file)
        Path(file.fullpath).unlink()
        new_folder.rmdir()

    def test_save_temp(self):
        fullpath = folder_nn3 / "only_variables.in"
        file = InputFile(fullpath)
        file.save(temp=True)
        self.assertTrue(Path(file.fullpath).is_file())
        try:
            self.assertTrue(Path(file.fullpath).is_file())
            # Check that the file is in the system temp directory
            temp_dir = Path(tempfile.gettempdir()).resolve()
            file_dir = Path(file.fullpath).resolve().parent
            self.assertTrue(str(file_dir).startswith(str(temp_dir)))
        finally:
            if Path(file.fullpath).exists():
                Path(file.fullpath).unlink()

    def test_text(self):
        fullpath_onlyvar = folder_nn3 / "only_variables.in"
        fullpath_example = folder_nn3 / "example.in"
        file = InputFile(fullpath_onlyvar)
        text = file.text

        new_file = InputFile(fullpath_example)
        new_file.text = text

        self.assertEqual(new_file.product, "nextnano3")
        self.assertEqual(new_file.text, text)
        self.assertEqual(new_file.lines, file.lines)
        self.assertEqual(new_file.variables["MAYUS"].name, "MAYUS")
        self.assertEqual(new_file.variables["MAYUS"].value, "TEXT")
        self.assertEqual(new_file.variables["MAYUS"].comment, "")
        self.assertEqual(new_file.variables["MAYUS"].text, "%MAYUS = TEXT")
        self.assertEqual(
            Path(new_file.save(file.fullpath, overwrite=False)),
            folder_nn3 / "only_variables_0.in",
        )
        Path(new_file.fullpath).unlink()

    def test_set_and_save(self):
        fullpath = folder_nn3 / "only_variables.in"
        file = InputFile(fullpath)
        file.set_variable(name="float", value=0.4)
        self.assertAlmostEqual(file.variables["float"].value, 0.4)

        self.addCleanup((folder_nn3 / "only_variables_0.in").unlink)
        file.save()
        self.assertTrue((folder_nn3 / "only_variables_0.in").is_file())

    def test_same_dir(self):
        current_directory = Path.cwd()
        self.addCleanup(os.chdir, current_directory)

        os.chdir(folder_nn3)

        path = "only_variables.in"
        file = InputFile(path)
        file.set_variable(name="float", value=0.3333)
        self.assertAlmostEqual(file.variables["float"].value, 0.3333)
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=".",
            exceptions=["only_variables.in"],
        )
        file.save()
        self.assertTrue(Path("only_variables_0.in").is_file())

    def test_content(self):
        fullpath = folder_nn3 / "only_variables.in"
        file = InputFile(fullpath)

        self.assertIsNone(file.content)


class Test_negf_classic(unittest.TestCase):
    def test_load(self):
        fullpath = folder_negf / "example.xml"
        file = InputFile(fullpath)

        self.assertEqual(file.product, "nextnano.NEGF_classic")

        self.assertEqual(len(file.variables.keys()), 4)

        self.assertEqual(file.variables["variable1"].name, "variable1")
        self.assertEqual(file.variables["variable1"].value, float(0.24))
        self.assertEqual(file.variables["variable1"].comment, "Some comment")
        self.assertEqual(file.variables["variable1"].unit, "")

        self.assertEqual(file.variables["variable2"].name, "variable2")
        self.assertEqual(file.variables["variable2"].value, int(0))
        self.assertEqual(file.variables["variable2"].comment, "Another comment")
        self.assertEqual(file.variables["variable2"].unit, "meV")

        self.assertEqual(file.variables["text_var"].name, "text_var")
        self.assertEqual(file.variables["text_var"].value, "some text")

        self.assertEqual(file.variables["ref_var"].value, "$(1-variable1)")

        fullpath = folder_negf / "virtual_file.xml"
        self.assertRaises(FileNotFoundError, InputFile, fullpath)

    def test_get_variables(self):
        fullpath = folder_negf / "example.xml"
        file = InputFile(fullpath)

        self.assertEqual(file.variables["variable1"], file.get_variable("variable1"))

        self.assertRaises(KeyError, file.get_variable, name="new_variable")

    def test_set_variables(self):
        fullpath = folder_negf / "example.xml"
        file = InputFile(fullpath)

        file.set_variable("variable1", 0.137, "test comment", "test unit^2")
        self.assertEqual(file.variables["variable1"].value, float(0.137))
        self.assertEqual(file.variables["variable1"].comment, "test comment")
        self.assertEqual(file.variables["variable1"].unit, "test unit^2")

        file.set_variable("text_var", "string variable test")
        self.assertEqual(file.variables["text_var"].value, "string variable test")

        file.set_variable("ref_var", "$variable1 - 0.1")
        self.assertEqual(file.variables["ref_var"].value, "$variable1 - 0.1")

        self.assertRaises(KeyError, file.set_variable, name="new_variable")

    def test_fullpath(self):
        fullpath = folder_negf / "example.xml"
        file = InputFile(fullpath)

        self.assertEqual(Path(file.fullpath), fullpath)
        self.assertEqual(
            Path(file.save(file.fullpath, overwrite=False)),
            folder_negf / "example_0.xml",
        )
        Path(file.fullpath).unlink()

    def test_config(self):
        fullpath = folder_negf / "example.xml"
        file = InputFile(fullpath)
        from nextnanopy import config

        for key, value in config.config["nextnano.NEGF_classic"].items():
            self.assertEqual(file.default_command_args[key], value)

    def test_save(self):
        fullpath = folder_negf / "example.xml"
        file = InputFile(fullpath)

        new_folder = folder_negf / "temp"
        new_file = new_folder / "example_copy.in"
        self.assertRaises(
            FileNotFoundError, file.save, new_file, overwrite=True, automkdir=False
        )
        self.assertEqual(Path(file.save(new_file, overwrite=True, automkdir=True)), new_file)
        Path(file.fullpath).unlink()
        new_folder.rmdir()

    def test_set_and_save(self):
        fullpath = folder_negf / "example.xml"
        file = InputFile(fullpath)
        file.set_variable(name="variable1", value=0.4)
        self.assertAlmostEqual(file.variables["variable1"].value, 0.4)

        self.addCleanup((folder_negf / "example_0.xml").unlink)
        file.save()
        self.assertTrue((folder_negf / "example_0.xml").is_file())

    def test_content(self):
        fullpath = folder_negf / "example.xml"
        file = InputFile(fullpath)

        self.assertIsNone(file.content)


class Test_negf(unittest.TestCase):
    # TODO: implement test for NEGF++ input file
    def test_load(self):
        fullpath = folder_negf / "Minimal_InputFile.negf"

        file = InputFile(fullpath)
        self.assertEqual(file.product, "nextnano.NEGF")

        self.assertEqual(len(file.variables.keys()), 1)
        self.assertEqual(file.variables["alloyComposition"].name, "alloyComposition")
        self.assertAlmostEqual(
            file.variables["alloyComposition"].value, float(0.15), delta=1e-9
        )
        self.assertEqual(
            file.variables["alloyComposition"].comment, "alloy composition"
        )

        fullpath = folder_nnp / "virtual_file.in"
        self.assertRaises(FileNotFoundError, InputFile, fullpath)

    def test_get_variables(self):
        fullpath = folder_negf / "Minimal_InputFile.negf"
        file = InputFile(fullpath)

        self.assertEqual(
            file.variables["alloyComposition"], file.get_variable("alloyComposition")
        )
        self.assertRaises(KeyError, file.get_variable, name="new_variable")

    def test_set_variables(self):
        fullpath = folder_negf / "Minimal_InputFile.negf"
        file = InputFile(fullpath)
        file.set_variable(
            "alloyComposition", 1e-5, "some comment", "some unit"
        )  # unit is not visible in nn++

        self.assertEqual(file.variables["alloyComposition"].name, "alloyComposition")
        self.assertEqual(file.variables["alloyComposition"].value, 1e-5)
        self.assertEqual(file.variables["alloyComposition"].comment, "some comment")
        self.assertEqual(
            file.variables["alloyComposition"].text,
            f"$alloyComposition = {str(1e-5)} # some comment",
        )
        self.assertEqual(
            file.lines[0], f"$alloyComposition = {str(1e-5)} # some comment"
        )

        file.set_variable("alloyComposition", value=1e-7)
        self.assertEqual(file.variables["alloyComposition"].name, "alloyComposition")
        self.assertEqual(file.variables["alloyComposition"].value, 1e-7)
        self.assertEqual(file.variables["alloyComposition"].comment, "some comment")
        self.assertEqual(
            file.variables["alloyComposition"].text,
            f"$alloyComposition = {str(1e-7)} # some comment",
        )
        self.assertEqual(
            file.lines[0], f"$alloyComposition = {str(1e-7)} # some comment"
        )

        file.set_variable("alloyComposition", comment="new comment")
        self.assertEqual(file.variables["alloyComposition"].name, "alloyComposition")
        self.assertEqual(file.variables["alloyComposition"].value, 1e-7)
        self.assertEqual(file.variables["alloyComposition"].comment, "new comment")
        self.assertEqual(
            file.variables["alloyComposition"].text,
            f"$alloyComposition = {str(1e-7)} # new comment",
        )
        self.assertEqual(
            file.lines[0], f"$alloyComposition = {str(1e-7)} # new comment"
        )

        file.set_variable("alloyComposition", value=0)
        self.assertEqual(file.variables["alloyComposition"].value, 0)

        self.assertRaises(KeyError, file.set_variable, name="new_variable")
        self.assertRaises(KeyError, file.set_variable, name="new_variable")

    def test_fullpath(self):
        fullpath = folder_negf / "Minimal_InputFile.negf"
        file = InputFile(fullpath)

        self.assertEqual(Path(file.fullpath), fullpath)
        self.assertEqual(
            Path(file.save(file.fullpath, overwrite=False)),
            folder_negf / "Minimal_InputFile_0.negf",
        )
        Path(file.fullpath).unlink()

    def test_config(self):
        fullpath = folder_negf / "Minimal_InputFile.negf"
        file = InputFile(fullpath)
        from nextnanopy import config

        for key, value in config.config["nextnano.NEGF"].items():
            self.assertEqual(file.default_command_args[key], value)

    def test_set_and_save(self):
        fullpath = folder_negf / "Minimal_InputFile.negf"
        file = InputFile(fullpath)
        file.set_variable(name="alloyComposition", value=0.4)
        self.assertAlmostEqual(file.variables["alloyComposition"].value, 0.4)

        self.addCleanup((folder_negf / "Minimal_InputFile_0.negf").unlink)
        file.save()
        self.assertTrue((folder_negf / "Minimal_InputFile_0.negf").is_file())

    def test_same_dir_saving(self):
        current_directory = Path.cwd()
        self.addCleanup(os.chdir, current_directory)

        os.chdir(folder_negf)

        path = "Minimal_InputFile.negf"
        file = InputFile(path)
        file.set_variable(name="alloyComposition", value=0.3333)
        self.assertAlmostEqual(file.variables["alloyComposition"].value, 0.3333)
        self.addCleanup(
            delete_files,
            "Minimal_InputFile",
            directory=".",
            exceptions=["Minimal_InputFile.negf"],
        )
        file.save()
        self.assertTrue(Path("Minimal_InputFile_0.negf").is_file())


class Test_msb(unittest.TestCase):
    def test_load(self):
        fullpath = folder_msb / "example.msb"

        input_file = InputFile(fullpath)
        print(input_file)
        print(input_file.text)
        print(type(input_file))
        self.assertEqual(input_file.product, "nextnano.MSB")
        self.assertEqual(len(input_file.variables.keys()), 1)

        fullpath = folder_msb / "virtual_file.msb"
        self.assertRaises(FileNotFoundError, InputFile, fullpath)

    def get_variables(self):
        fullpath = folder_msb / "example.msb"
        input_file = InputFile(fullpath)

        self.assertEqual(
            input_file.variables["Well"],
            input_file.get_variable("Well"),
        )
        self.assertRaises(KeyError, input_file.get_variable, name="new_variable")

        self.assertEqual(
            input_file.variables["Well"].value,
            0.14
        )

    def test_set_variable(self):
        fullpath = folder_msb / "example.msb"
        input_file = InputFile(fullpath)

        input_file.set_variable("Well", 0.2, "Updated well width", "nm")
        self.assertEqual(input_file.variables["Well"].value, 0.2)
        self.assertEqual(input_file.variables["Well"].comment, "Updated well width")
        self.assertEqual(input_file.variables["Well"].unit, "nm")

    def test_set_and_save(self):
        fullpath = folder_msb / "example.msb"
        input_file = InputFile(fullpath)

        input_file.set_variable(name="Well", value=0.25)
        self.assertAlmostEqual(input_file.variables["Well"].value, 0.25)
        self.addCleanup((folder_msb / "example_0.msb").unlink)
        input_file.save()
        self.assertTrue((folder_msb / "example_0.msb").is_file())
        input_file_1 = InputFile(folder_msb / "example_0.msb")
        self.assertAlmostEqual(input_file_1.variables["Well"].value, 0.25)


class TestInputFile(unittest.TestCase):

    def test_access_by_index(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)
        for key, value in file.variables.items():
            self.assertEqual(file[key], value)

    def test_for_loop(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)
        for i, fi in enumerate(file):
            self.assertEqual(file.variables[i], fi)

    def test_fullpath(self):
        fullpath = folder_nnp / "only_variables.in"
        file = InputFile(fullpath)
        self.assertEqual(Path(file.fullpath), fullpath)
        self.assertEqual(file.filename, "only_variables.in")
        self.assertEqual(file.filename_only, "only_variables")
        self.assertEqual(Path(file.folder_input), folder_nnp)
        self.assertEqual(file.execute_info, {})

        file.filename = "new_name.in"
        self.assertEqual(file.filename, "new_name.in")
        self.assertEqual(file.filename_only, "new_name")
        self.assertEqual(Path(file.fullpath), folder_nnp / "new_name.in")
        self.assertEqual(Path(file.folder_input), folder_nnp)

        file.filename_only = "another_filename"
        self.assertEqual(file.filename, "another_filename.in")
        self.assertEqual(file.filename_only, "another_filename")
        self.assertEqual(Path(file.fullpath), folder_nnp / "another_filename.in")
        self.assertEqual(Path(file.folder_input), folder_nnp)

        npath = Path("new") / "folder"
        file.folder_input = npath
        self.assertEqual(file.filename, "another_filename.in")
        self.assertEqual(file.filename_only, "another_filename")
        self.assertEqual(Path(file.fullpath), npath / "another_filename.in")
        self.assertEqual(Path(file.folder_input), npath)

        npath = Path("random") / "path"
        file.execute_info["outputdirectory"] = npath
        self.assertEqual(Path(file.folder_output), npath)


class TestSweep(unittest.TestCase):
    def test_init(self):
        self.assertRaises(TypeError, Sweep)
        sweep = Sweep({})
        self.assertEqual(sweep.fullpath, None)
        self.assertEqual(sweep.var_sweep, {})
        self.assertRaises(ValueError, Sweep, {"Name": "some_name"})

    def test_nnp_init(self):
        fullpath = folder_nnp / "only_variables.in"
        self.assertRaises(ValueError, Sweep, {"Name": "some_name"}, fullpath)

        sweep = Sweep({}, fullpath)
        self.assertEqual(Path(sweep.fullpath), fullpath)
        self.assertEqual(sweep.var_sweep, {})
        self.assertFalse(sweep.input_files)
        self.assertFalse(sweep.sweep_output_directory)

        self.assertRaises(ValueError, Sweep, {"float": 1})
        self.assertRaises(TypeError, Sweep, {"float": 1}, fullpath)

        sweep = Sweep({"float": [1, 2, 5]}, fullpath)
        self.assertEqual(Path(sweep.fullpath), fullpath)
        self.assertEqual(sweep.var_sweep["float"], [1, 2, 5])

        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nnp,
            exceptions=["only_variables.in"],
        )

    def test_nnp_mkdir(self):
        fullpath = folder_nnp / "only_variables.in"
        sweep = Sweep({}, fullpath=fullpath)
        sweep.config.set("nextnano++", "outputdirectory", r"tests//outputs")
        created_directory = sweep.mk_dir(overwrite=True)
        self.assertTrue(
            Path(created_directory).samefile(
                Path("tests") / "outputs" / "only_variables_sweep"
            )
        )
        self.assertTrue(
            (Path("tests") / "outputs" / "only_variables_sweep").is_dir()
        )

        self.addCleanup((Path("tests") / "outputs" / "only_variables_sweep").rmdir)
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nnp,
            exceptions=["only_variables.in"],
        )

    def test_nnp_mkdir_specify(self):
        fullpath = folder_nnp / "only_variables.in"
        sweep = Sweep({}, fullpath=fullpath)
        created_directory = sweep.mk_dir(overwrite=True, output_directory=r"tests//outputs")
        self.assertTrue(
            Path(created_directory).samefile(
                Path("tests") / "outputs" / "only_variables_sweep"
            )
        )
        self.assertTrue(
            (Path("tests") / "outputs" / "only_variables_sweep").is_dir()
        )

        self.addCleanup((Path("tests") / "outputs" / "only_variables_sweep").rmdir)
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nnp,
            exceptions=["only_variables.in"],
        )

    def test_nnp_save(self):
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nnp,
            exceptions=["only_variables.in"],
        )
        fullpath = folder_nnp / "only_variables.in"
        sweep = Sweep({"float": [1, 2], "str": ["test1", "test2"]}, fullpath)
        sweep.save_sweep()

        files_with_names = [
            p.name for p in folder_nnp.iterdir() if "only_variables" in p.name
        ]
        self.assertEqual(len(files_with_names), 5)
        self.assertTrue(
            (folder_nnp / "only_variables__float_2_str_test1_.in").is_file()
        )

    def test_nnp_save_temp(self):
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nnp,
            exceptions=["only_variables.in"],
        )

        fullpath = folder_nnp / "only_variables.in"
        sweep = Sweep({"float": [1, 2], "str": ["test1", "test2"]}, fullpath)
        sweep.save_sweep(temp=True)

        files_with_names = [
            file for file in os.listdir(folder_nnp) if "only_variables" in file
        ]
        self.assertEqual(len(files_with_names), 1)
        self.assertFalse(
            (folder_nnp / "only_variables__float_2_str_test1_.in").is_file()
        )


    # nn3 section
    def test_nn3_init(self):
        fullpath = folder_nn3 / "only_variables.in"
        self.assertRaises(ValueError, Sweep, {"Name": "some_name"}, fullpath)

        sweep = Sweep({}, fullpath)
        self.assertEqual(Path(sweep.fullpath), fullpath)
        self.assertEqual(sweep.var_sweep, {})
        self.assertFalse(sweep.input_files)
        self.assertFalse(sweep.sweep_output_directory)

        self.assertRaises(ValueError, Sweep, {"float": 1})
        self.assertRaises(TypeError, Sweep, {"float": 1}, fullpath)

        sweep = Sweep({"float": [1, 2, 5]}, fullpath)
        self.assertEqual(Path(sweep.fullpath), fullpath)
        self.assertEqual(sweep.var_sweep["float"], [1, 2, 5])

        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nn3,
            exceptions=["only_variables.in"],
        )

    def test_nn3_mkdir(self):
        fullpath = folder_nn3 / "only_variables.in"
        sweep = Sweep({}, fullpath=fullpath)
        sweep.config.set("nextnano3", "outputdirectory", r"tests//outputs")
        created_directory = sweep.mk_dir(overwrite=True)
        self.assertTrue(
            Path(created_directory).samefile(
                Path("tests") / "outputs" / "only_variables_sweep"
            )
        )
        self.assertTrue(
            (Path("tests") / "outputs" / "only_variables_sweep").is_dir()
        )

        self.addCleanup((Path("tests") / "outputs" / "only_variables_sweep").rmdir)
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nn3,
            exceptions=["only_variables.in"],
        )

    def test_nn3_save(self):
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nn3,
            exceptions=["only_variables.in"],
        )
        fullpath = folder_nn3 / "only_variables.in"
        sweep = Sweep({"float": [1, 2], "str": ["test1", "test2"]}, fullpath)
        sweep.save_sweep()

        files_with_names = [
            p.name for p in folder_nn3.iterdir() if "only_variables" in p.name
        ]
        self.assertEqual(len(files_with_names), 5)
        self.assertTrue(
            (folder_nn3 / "only_variables__float_2_str_test1_.in").is_file()
        )

    # TODO test parallel sweeps with and without convergenceCheck
    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        sweep_dir = Path("tests") / "outputs" / "only_variables_sweep"
        if sweep_dir.is_dir():
            sweep_dir.rmdir()
        delete_files(
            "only_variables", directory=folder_nnp, exceptions=["only_variables.in"]
        )
        delete_files(
            "only_variables", directory=folder_nn3, exceptions=["only_variables.in"]
        )

    def test_conditional_sweep(self):
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nnp,
            exceptions=["only_variables.in"],
        )
        def condition(combination):
            return combination[0] > 0.2

        fullpath = folder_nnp / "only_variables.in"
        sweep = Sweep(
            {"float": [0.1, 0.2, 0.3, 0.4, 0.5]},
            fullpath,
        )

        sweep.save_sweep(variables_comb_screen_fn=condition)

        self.assertEqual(len(sweep.input_files), 3)

        for combination in sweep.sweep_infodict.values():
            combination = list(combination.values())
            combination[0] > 0.2

    def test_conditional_sweep_multivar(self):
        self.addCleanup(
            delete_files,
            "only_variables",
            directory=folder_nnp,
            exceptions=["only_variables.in"],
        )
        def condition(combination):
            return combination[1] > combination[0]
        fullpath = folder_nnp / "only_variables.in"
        sweep = Sweep(
            {"float": [0.5, 1.5, 2.5], "int": [1, 2, 3]},
            fullpath,
        )
        sweep.save_sweep(variables_comb_screen_fn=condition)

        self.assertEqual(len(sweep.input_files), 6)

        for combination in sweep.sweep_infodict.values():
            combination = list(combination.values())
            self.assertTrue(combination[1] > combination[0])


if __name__ == "__main__":
    unittest.main()

    file = folder_nnp / "only_variables.in"
    fi = InputFile(file)
    text = fi.text



import unittest
from nextnanopy.utils.misc import *


class TestMisc(unittest.TestCase):
    def test_find_name_max_idx(self):
        self.assertEqual(find_unused_name('ex.in', [''], extension='.in', max_idx=True), 'ex_0.in')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.in'], extension='.in', max_idx=True), 'ex_1.in')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.in', 'ex_1.in'], extension='.in', max_idx=True), 'ex_2.in')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.in', 'ex_2.in'], extension='.in', max_idx=True), 'ex_3.in')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.in', 'ex_2.in'], extension='.txt', max_idx=True),
                         'ex.in_0.txt')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.txt', 'ex_2.txt'], extension='.in', max_idx=True), 'ex_0.in')

        self.assertEqual(find_unused_name('ex_0.in', [''], extension='.in', max_idx=True), 'ex_0.in')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.in'], extension='.in', max_idx=True), 'ex_1.in')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.in', 'ex_1.in'], extension='.in', max_idx=True), 'ex_2.in')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.in', 'ex_2.in'], extension='.in', max_idx=True), 'ex_3.in')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.in', 'ex_2.in'], extension='.txt', max_idx=True),
                         'ex_0.txt')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.txt', 'ex_2.txt'], extension='.in', max_idx=True),
                         'ex_0.in')

        self.assertEqual(find_unused_name('ex_0d_0.in', ['ex_0d_0.in'], extension='.in', max_idx=True), 'ex_0d_1.in')
        self.assertEqual(find_unused_name('ex_0d_0.in', ['ex_0d_0.in', 'ex_0d_2.in'], extension='.in', max_idx=True),
                         'ex_0d_3.in')

    def test_find_name_min_idx(self):
        self.assertEqual(find_unused_name('ex.in', [''], extension='.in', max_idx=False), 'ex_0.in')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.in'], extension='.in', max_idx=False), 'ex_1.in')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.in', 'ex_1.in'], extension='.in', max_idx=False), 'ex_2.in')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.in', 'ex_2.in'], extension='.in', max_idx=False), 'ex_1.in')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.in', 'ex_2.in'], extension='.txt', max_idx=False),
                         'ex.in_0.txt')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.txt', 'ex_2.txt'], extension='.in', max_idx=False), 'ex_0.in')
        self.assertEqual(find_unused_name('ex.in', ['ex_0.txt', 'ex_2.in'], extension='.in', max_idx=False), 'ex_0.in')

        self.assertEqual(find_unused_name('ex_0.in', [''], extension='.in', max_idx=False), 'ex_0.in')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.in'], extension='.in', max_idx=False), 'ex_1.in')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.in', 'ex_1.in'], extension='.in', max_idx=False), 'ex_2.in')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.in', 'ex_2.in'], extension='.in', max_idx=False), 'ex_1.in')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.in', 'ex_2.in'], extension='.txt', max_idx=False),
                         'ex_0.txt')
        self.assertEqual(find_unused_name('ex_0.in', ['ex_0.txt', 'ex_2.txt'], extension='.in', max_idx=False),
                         'ex_0.in')
        self.assertEqual(find_unused_name('ex_0d_0.in', ['ex_0d_0.in'], extension='.in', max_idx=False), 'ex_0d_1.in')
        self.assertEqual(find_unused_name('ex_0d_0.in', ['ex_0d_0.in', 'ex_0d_2.in'], extension='.in', max_idx=False),
                         'ex_0d_1.in')

if __name__ == '__main__':
    unittest.main()



from collections import OrderedDict


class DictList(OrderedDict):
    """
        This is a customized collections.OrderedDict
        It is a mixture between a dictionary and a list because it allows to
        access the values via keys or integer indexes.
        Common usage:
            d = DictList(a=3, b='t')
            d[0] = 3
            d['a'] = 3

        It has all the methods and attributes from a dictionary like .keys(), .values(), .items()

        Moreover, it supports iterations like a list:
        for value in d:
            print(value)
        >>> 3
        >>> 't'


    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def _idxs(self):
        return OrderedDict(enumerate(self.keys()))

    def get_indx(self, key):
        for i, _key in enumerate(self.keys()):
            if _key == key:
                return i
        raise KeyError('No such key in the DictList')

    def __getitem__(self, key):
        if isinstance(key, int):
            if key < 0:
                key = len(self) + key
            key = self._idxs[key]
        item = super().__getitem__(key)
        return item

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        cname = self.__class__.__name__
        args = [f"(index: {idx} - key: '{key}' - {value})" for idx, key, value in
                zip(self._idxs.keys(), self.keys(), self.values())]
        args = ',\n'.join(args)
        return f"{cname}([\n{args}\n])"

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        try:
            result = self.__getitem__(self._iter_index)
        except (IndexError, KeyError):
            raise StopIteration
        self._iter_index += 1
        return result
